from enum import Enum
from typing import List, Dict
from uuid import uuid4
from semver import VersionInfo

from cyst.api.logic.exploit import ExploitCategory, ExploitLocality, ExploitParameterType, VulnerableService, Exploit, ExploitParameter


class ExploitParameterImpl(ExploitParameter):
    # Immutable parameters are further specifications of an exploit, which can be used by the system, but cannot
    # be altered by the attacker (although they can be queried [and directly accessed by that's another story])
    def __init__(self, exploit_type: ExploitParameterType, value: str = "", immutable: bool = False):
        self._type = exploit_type
        self._value = value
        self._immutable = immutable

    @property
    def type(self) -> ExploitParameterType:
        return self._type

    @property
    def value(self) -> str:
        return self._value

    def set_value(self, value) -> None:
        # Failure to bypass immutability is silently ignored
        if not self._immutable:
            self._value = value


class VulnerableServiceImpl(VulnerableService):
    def __init__(self, id: str, min_version: str = "0.0.0", max_version: str = "0.0.0"):
        self._id = id

        if max_version == "0.0.0":
            max_version = min_version

        self._min_ver = VersionInfo.parse(min_version)
        self._max_ver = VersionInfo.parse(max_version)

        if self._min_ver > self._max_ver:
            raise Exception("Version mismatch, minimal version larger than maximal")

    @property
    def id(self):
        return self._id

    @property
    def min_version(self):
        return self._min_ver

    @property
    def max_version(self):
        return self._max_ver


class ExploitImpl(Exploit):
    def __init__(self, id: str = "", services: List[VulnerableService] = None, locality:
                 ExploitLocality = ExploitLocality.NONE, category: ExploitCategory = ExploitCategory.NONE,
                 *parameters: ExploitParameter) -> None:
        if not id:
            self._id = str(uuid4())
        else:
            self._id = id

        self._services = {}
        if services:
            for service in services:
                self._services[service.id] = service

        self._locality = locality
        self._category = category

        self._parameters = {}
        if parameters:
            for param in parameters:
                self._parameters[param.type] = param

    @property
    def id(self) -> str:
        return self._id

    @property
    def services(self) -> Dict[str, VulnerableService]:
        return self._services

    @property
    def locality(self) -> ExploitLocality:
        return self._locality

    @property
    def category(self) -> ExploitCategory:
        return self._category

    @property
    def parameters(self) -> Dict[ExploitParameterType, ExploitParameter]:
        return self._parameters



from abc import abstractmethod, ABC
from netaddr import IPAddress
from typing import List, Optional, Tuple


class Session(ABC):
    """
    A session represents a virtual connection between two nodes in the infrastructure. The session ignores the routing
    limitations imposed by router configuration.
    """
    @property
    @abstractmethod
    def owner(self) -> str:
        """
        Returns the owner of the session. This method is being slowly phased away.

        :rtype: str
        """

    @property
    @abstractmethod
    def id(self) -> str:
        """
        Returns a unique identifier of the session.

        :rtype: str
        """

    @property
    @abstractmethod
    def parent(self) -> Optional['Session']:
        """
        Returns the parent session of this session (if there is one). Sessions can be chained together and a message
        traverses this chain sequentially.

        :rtype: Optional[Session]
        """

    @property
    @abstractmethod
    def path(self) -> List[Tuple[Optional[IPAddress], Optional[IPAddress]]]:
        """
        Returns a path trough the infrastructure, on which the session lies. The path is modelled as a set of "hops",
        i.e., tuples of IP addresses representing a source and destination of one connection. Note that a destination
        of preceding hop does not have to be the same as the source of the next hop. These can often differ, especially
        when a session is crossing between different networks.

        :rtype: List[Tuple[Optional[IPAddress], Optional[IPAddress]]]
        """

    @property
    @abstractmethod
    def end(self) -> Tuple[IPAddress, str]:
        """
        Returns the ultimate destination of the session. The result includes both the IP address of the node, and the
        service the session is linked to.

        :rtype: Tuple[IPAddress, str]
        """

    @property
    @abstractmethod
    def start(self) -> Tuple[IPAddress, str]:
        """
        Returns the start of the session. If the session has a parent, then a start of the parent is returned.
        The result includes both the IP address of the node, and the service the session is linked to.

        :rtype: Tuple[IPAddress, str]
        """

    @property
    @abstractmethod
    def enabled(self) -> bool:
        """
        Returns whether the session is enabled. Messages can't be send in disabled sessions.

        :rtype: bool
        """



from netaddr import IPAddress
from typing import List, Union, Optional, Tuple, Dict

from cyst.api.host.service import Service, ActiveService
from cyst.api.environment.message import MessageType
from cyst.api.network.node import Node
from cyst.api.network.elements import Interface

from cyst.platform.network.elements import InterfaceImpl
from cyst.platform.host.service import ServiceImpl


class NodeImpl(Node):
    def __init__(self, id: str, type: str = "Node", ip: Union[str, IPAddress] = "", mask: str = "", shell: Service = None):
        self._id: str = id
        self._type: str = type
        self._interfaces: List[InterfaceImpl] = []
        self._services: Dict[str, ServiceImpl] = {}
        self._ip: Optional[IPAddress] = None
        if ip:
            self._interfaces.append(InterfaceImpl(ip, mask))
        self._shell = shell
        self._traffic_processors: List[ActiveService] = []

    @property
    def id(self) -> str:
        return self._id

    # ------------------------------------------------------------------------------------------------------------------
    # Node interface
    @property
    def type(self) -> str:
        return self._type

    @property
    def interfaces(self) -> List[Interface]:
        return self._interfaces

    @property
    def services(self) -> Dict[str, Service]:
        return self._services

    # Gateway returns both the IP address of the gateway and the port index
    def gateway(self, ip: Union[str, IPAddress] = "") -> Optional[Tuple[IPAddress, int]]:
        # If no IP is specified the the first gateway is used as a default gateway
        if not self._interfaces:
            return None

        # Explicit query for default gateway
        if not ip:
            return self._interfaces[0].gateway, 0

        # Checking all available routes for exact one
        for iface in self._interfaces:
            if iface.routes(ip):
                return iface.gateway, iface.index

        # Using a default one
        return self._interfaces[0].gateway, 0
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def ips(self) -> List[IPAddress]:
        # A set is used because a router can have multiple interfaces with the same IP address
        result = set()
        for i in self.interfaces:
            result.add(i.ip)
        return list(result)

    def add_interface(self, i: InterfaceImpl) -> int:
        # TODO Currently there is no control of interface overlaps. Question is whether it matters...
        self._interfaces.append(i)
        index = len(self._interfaces) - 1
        i.set_index(index)
        return index

    async def process_message(self, message, delay: float) -> int:
        if message.type == MessageType.ACK:
            return 0

        print("Processing message at node {}. {}".format(self.id, message))
        return 0

    @property
    def shell(self) -> Optional[Service]:
        return self._shell

    def set_shell(self, value: Service) -> None:
        self._shell = value

    def add_service(self, service: Service) -> None:
        s = ServiceImpl.cast_from(service)
        # Passive services are identified by their type, active by their name,
        if s.passive:
            self._services[s.type] = s
        else:
            self._services[s.name] = s

        s.set_node(self._id)
        # TODO: create a mechanism for reasonable running of active services
        # Initiate active services
        # if not s.passive:
        #    s.active_service.run()

    def remove_service(self, service: Service) -> None:
        s = ServiceImpl.cast_from(service)
        # Passive services are identified by their type, active by their name,
        if s.passive and s.type in self._services:
            del self._services[s.type]
        elif not s.passive and s.name in self._services:
            del self._services[s.name]

    def add_traffic_processor(self, value: ActiveService) -> None:
        self._traffic_processors.append(value)

    @property
    def traffic_processors(self) -> List[ActiveService]:
        return self._traffic_processors

    def __str__(self) -> str:
        return "Node: [Shell: {}, Services: {}, Interfaces: {}]".format(self.shell, self.services, self.interfaces)

    @staticmethod
    def cast_from(o: Node) -> 'NodeImpl':
        if isinstance(o, NodeImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the Node interface")



from copy import deepcopy
from typing import Dict, Union, List

from cyst.api.environment.data_model import ActionModel
from cyst.api.environment.message import Message, Signal
from cyst.api.environment.stats import Statistics
from cyst.api.environment.stores import DataStore, DataStoreDescription


class DataStoreMemory(DataStore):
    def __init__(self, run_id: str, params: Dict[str, str]):
        self._run_id = run_id
        self._memory: Dict[str, Union[None, List, Statistics]] = {
            "actions": [],
            "messages": [],
            "statistics": None,
            "signals": []
        }

    def add_action(self, *action: ActionModel) -> None:
        for a in action:
            self._memory["actions"].append(deepcopy(a))

    def add_message(self, *message: Message) -> None:
        for m in message:
            self._memory["messages"].append(deepcopy(m))

    def add_statistics(self, statistics: Statistics) -> None:
        self._memory["statistics"] = statistics

    def add_signal(self, *signal: Signal) -> None:
        for s in signal:
            self._memory["signals"].append(deepcopy(s))

    @property
    def memory(self) -> Dict[str, Union[None, List, Statistics]]:
        return self._memory

def create_data_store_memory(run_id: str, params: Dict[str, str]) -> DataStore:
    return DataStoreMemory(run_id, params)


data_store_memory_description = DataStoreDescription(
    backend="memory",
    description="A memory-based data store backend. Due to a limited options to retrieve the system data from a data "
                "store, it is useful mostly when a only user data handling is required.",
    creation_fn=create_data_store_memory
)



from typing import List, Union, Optional, Tuple

from cyst.api.environment.configuration import EnvironmentConfiguration
from cyst.api.environment.policy import EnvironmentPolicy
from cyst.api.host.service import Service
from cyst.api.logic.access import AccessLevel, AccessScheme, Authorization
from cyst.api.network.node import Node

from cyst.platform.environment.configurator import Configurator
from cyst.platform.host.service import PassiveServiceImpl
from cyst.platform.logic.access import AuthorizationImpl, AccessSchemeImpl
from cyst.platform.network.node import NodeImpl


class Policy(EnvironmentPolicy):

    def __init__(self, configuration: Configurator):
        self._config = configuration

    def create_authorization(self, identity: str, nodes: List[Union[str, Node]], services: List[Union[str, Service]],
                             access_level: AccessLevel, id: str, token: Optional[str] = None) -> Optional[Authorization]:

        if not nodes or not services:
            return None

        auth = AuthorizationImpl(identity, list(map(lambda node: node if isinstance(node, str) else node.id, nodes)), #MYPY: add cast to nodeimpl Node itself does not have ID, not IMPL has?
                                 list(map(lambda service: service if isinstance(service, str) else service.name, services)),
                                 access_level, id, token)

        if len(nodes) > 1 or services == ["*"]: # federated, temporary solution
            return auth

        actual_node = nodes[0] if isinstance(nodes[0], Node) else self._config.get_object_by_id(nodes[0], Node) #type: ignore

        # if nodes is a [str], is it id, ip or else??

        for service in services:
            actual_service = service if isinstance(service, Service)\
                else actual_node.services.get(service)
            if isinstance(actual_service, PassiveServiceImpl):
                actual_service.add_active_authorization(auth)

        return auth


    def get_authorizations(self, node: Union[str, Node], service: str, access_level: AccessLevel = AccessLevel.NONE) -> \
    List[Authorization]:
        """ This only return the Authorization templates"""

        actual_node = node if isinstance(node, Node) else self._config.get_object_by_id(node, Node) #type: ignore


        actual_service = actual_node.services.get(service)

        auths = []

        if isinstance(actual_service, PassiveServiceImpl):
            for scheme in map(lambda a_sch: AccessSchemeImpl.cast_from(a_sch), actual_service.access_schemes):
                auths.extend(scheme.authorizations)

        return auths



    def decide(self, node: Union[str, Node], service: str, access_level: AccessLevel, authorization: Authorization) -> \
    Tuple[bool, str]:

        actual_node = node if isinstance(node, Node) else self._config.get_object_by_id(node, Node)

        actual_service = actual_node.services.get(service)

        if not actual_service:
            return False, "Service does not exist on node."

        retval = None

        if isinstance(actual_service, PassiveServiceImpl):
            retval = actual_service.assess_authorization(authorization, access_level,
                                                         NodeImpl.cast_from(actual_node).id, service)

        if not retval:
            return False, "Invalid service type."

        return retval


    def get_schemas(self, node: Union[str, Node], service: str) -> List[AccessScheme]:
        actual_node = node if isinstance(node, Node) else self._config.get_object_by_id(node, Node)
        actual_service = actual_node.services.get(service)

        return actual_service._access_schemes if isinstance(actual_service, PassiveServiceImpl) else []


    def get_nodes(self, authorization: Authorization) -> List[str]:
        return AuthorizationImpl.cast_from(authorization).nodes

    def get_services(self, authorization: Authorization) -> List[str]:
        return AuthorizationImpl.cast_from(authorization).services

    def get_access_level(self, authorization: Authorization) -> AccessLevel:
        return AuthorizationImpl.cast_from(authorization).access_level



from copy import copy
from dataclasses import dataclass, field
from enum import IntEnum
from typing import List, Optional, Union, Tuple
from uuid import uuid4

from netaddr import IPAddress
from serde import serialize, coerce
from serde.compat import typename

from cyst.api.configuration.configuration import ConfigItem
from cyst.api.logic.access import AccessLevel, AuthenticationTokenSecurity, AuthenticationTokenType, \
                                  AuthenticationProviderType


@serialize(type_check=coerce)
@dataclass
class AuthorizationConfig(ConfigItem):
    """ Configuration of a local authorization.

    This configuration is used as a template to produce authorization tokens after successful authentication.

    :param identity: An identity, who this authorization relates to.
    :type identity: str

    :param access_level: An access level of this particular authorization
    :type access_level: AccessLevel
    """
    identity: str
    access_level: AccessLevel
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__authorization"
    id: str = ""


@serialize(type_check=coerce)
@dataclass
class FederatedAuthorizationConfig(ConfigItem):
    """ Configuration of a federated authorization.

    Unlike local authorization a federated authorization can span multiple services and nodes.

    This configuration is used as a template to produce authorization tokens after successful authentication.

    :param identity: An identity, who this authorization relates to.
    :type identity: str

    :param access_level: An access level of this particular authorization
    :type access_level: AccessLevel

    :param nodes: A list of node ids this authorization applies to.
    :type nodes: List[str]

    :param services: A list of service ids this authorization applies to.
    :type services: List[str]
    """
    identity: str
    access_level: AccessLevel
    nodes: List[str]
    services: List[str]
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__authorization"
    id: str = ""


class AuthorizationDomainType(IntEnum):
    """ Specification of an authorization domain type.

    :LOCAL: Local domain (confined to one node and service)
    :FEDERATED: Federated domain (can span multiple nodes and services)
    """
    LOCAL = 0,
    FEDERATED = 1


@serialize(type_check=coerce)
@dataclass
class AuthorizationDomainConfig(ConfigItem):
    """ Configuration of an authorization domain.

    An authorization domain represents a collection of authorizations, which can then be associated with access scheme.

    :param type: A type of the domain.
    :type type: AuthorizationDomainType

    :param authorizations: A list of authorization configurations
    :type authorizations: List[Union[AuthorizationConfig, FederatedAuthorizationConfig]]
    """
    type: AuthorizationDomainType
    authorizations: List[Union[AuthorizationConfig, FederatedAuthorizationConfig]]
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__authorization_domain"
    id: str = ""


@serialize(type_check=coerce)
@dataclass
class AuthenticationProviderConfig(ConfigItem):
    """ Configuration of an authentication provider

    Authentication provider represents an authentication mechanism that can be employed in services via the access
    scheme mechanism.

    :param provider_type: The type of authentication provider.
    :type provider_type: AuthenticationProviderType

    :param token_type: The type of tokens that are employed by this authentication provider.
    :type token_type: AuthenticationTokenType

    :param token_security: Security mechanism applied to stored tokens.
    :type token_security: AuthenticationTokenSecurity

    :param ip: An optional IP address, which is intended for remote or federated providers. It represents an IP address
        where this provider can be accessed.
    :type ip: Optional[IPAddress]
    """
    provider_type: AuthenticationProviderType
    token_type: AuthenticationTokenType
    token_security: AuthenticationTokenSecurity
    ip: Optional[IPAddress] = field(default=None, metadata={
        'serde_serializer': lambda x: {"cls_type": typename(type(x)), "value": str(x)},
    })
    timeout: int = 0
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__authentication_provider"
    id: str = ""


@serialize(type_check=coerce)
@dataclass
class AccessSchemeConfig(ConfigItem):
    """ Configuration of an access scheme.

    An access scheme is a combination of authentication providers, which use a supplied authorization domain. An access
    scheme provides means to describe multiple authentication scheme within one service or multi-factor authentication.

    Example:

    .. code-block:: python

        PassiveServiceConfig(
            ...
            authentication_providers=[
                AuthenticationProviderConfig(
                    provider_type=AuthenticationProviderType.LOCAL,
                    token_type=AuthenticationTokenType.PASSWORD,
                    token_security=AuthenticationTokenSecurity.SEALED,
                    timeout=30,
                    id="openssh_local_pwd_auth"
                )
            ],
            access_schemes=[AccessSchemeConfig(
                authentication_providers=["openssh_local_pwd_auth"],
                authorization_domain=AuthorizationDomainConfig(
                    type=AuthorizationDomainType.LOCAL,
                    authorizations=[
                        AuthorizationConfig("user1", AccessLevel.LIMITED, id="ssh_auth_1"),
                        AuthorizationConfig("user2", AccessLevel.LIMITED, id="ssh_auth_2"),
                        AuthorizationConfig("root", AccessLevel.ELEVATED)
                    ]
                )
            )],
            ...
        )

    :param authentication_providers: A list of authentication providers or their ids.
    :type authentication_providers: List[Union[AuthenticationProviderConfig, str]]

    :param authorization_domain: A domain from which authorization tokens are created after successful authentication.
    :type authorization_domain: Union[AuthorizationDomainConfig, str]
    """
    authentication_providers: List[Union[AuthenticationProviderConfig, str]]
    authorization_domain: Union[AuthorizationDomainConfig, str]
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__access_scheme"
    id: str = ""



from dataclasses import dataclass, field
from semver import VersionInfo
from typing import List, Union, Dict, Optional
from uuid import uuid4
from serde import serialize, coerce

from cyst.api.configuration.configuration import ConfigItem
from cyst.api.logic.exploit import ExploitParameterType, ExploitCategory, ExploitLocality


@serialize(type_check=coerce)
@dataclass
class VulnerableServiceConfig(ConfigItem):
    """ Configuration of vulnerable service.

    This represents a version range of an exploitable service and is used only by exploits and exploit evaluation
    mechanisms.

    :param service: A name of the service. It is an equivalent of PassiveService/PassiveServiceConfig name.
    :type service: str

    :param min_version: A minimum version of the service that is vulnerable. Even though the type supports any string,
        using anything else than string representation of a semantic version will lead to problems.
    :type min_version: Union[VersionInfo, str]

    :param max_version: A maximum version of the service that is vulnerable. Even though the type supports any string,
        using anything else than string representation of a semantic version will lead to problems.
    :type max_version: Union[VersionInfo, str]
    """
    service: str
    min_version: Union[VersionInfo, str]
    max_version: Union[VersionInfo, str] = field(default="0.0.0")
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__vulnerable_service"
    id: str = ""


@serialize(type_check=coerce)
@dataclass
class ExploitParameterConfig(ConfigItem):
    """ Configuration of exploit parameter.

    Exploit parameter either represents a specification of an exploit, which can't be modified by a user, but is
    taken into account by exploit evaluation mechanisms (e.g., if the impact is one user or all users), or provides a
    mean for a user to supply additional information necessary for full execution of the exploit (e.g., identity of
    a user to impersonate).

    Example:

    .. code-block:: python

        parameters=[
            ExploitParameterConfig(ExploitParameterType.IDENTITY, immutable=False),
            ExploitParameterConfig(ExploitParameterType.ENABLE_ELEVATED_ACCESS, "FALSE", immutable=True)
        ],

    :param type: A type of the exploit parameter.
    :type type: ExploitParameterType

    :param value: A value of the parameter.
    :type value: Optional[str]

    :param immutable: A flag indicating, whether a user can change the value of the parameter.
    :type immutable: bool
    """
    type: ExploitParameterType
    value: Optional[str] = None
    immutable: bool = True
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__exploit_parameter"
    id: str = ""


@serialize(type_check=coerce)
@dataclass
class ExploitConfig(ConfigItem):
    """ Configuration of a working exploit.

    If a definition of exploit exists, then a services it refers to are vulnerable, as long as their version match.
    Patching the service and bumping a version over the max vulnerable version is the only countermeasure.

    Example:

    .. code-block:: python

        ExploitConfig(
            services=[
                VulnerableServiceConfig(
                    name="bash",
                    min_version="5.0.0",
                    max_version="5.1.0"
                )
            ],
            locality=ExploitLocality.LOCAL,
            category=ExploitCategory.AUTH_MANIPULATION,
            parameters=[
                ExploitParameterConfig(
                    type=ExploitParameterType.ENABLE_ELEVATED_ACCESS,
                    value="TRUE",
                    immutable=True
                )
            ],
            id="bash_root_exploit"
        )

    :param services: A list of services that are vulnerable to this exploit.
    :type services: List[Union[VulnerableServiceConfig, str]]

    :param locality: Determines if the exploit can be used locally or remotely.
    :type locality: ExploitLocality

    :param category: Determines the category of the exploit.
    :type category: ExploitCategory

    :param parameters: An optional list of exploit parameters.
    :type parameters: Optional[List[ExploitParameterConfig]]
    """
    services: List[Union[VulnerableServiceConfig, str]]
    locality: ExploitLocality
    category: ExploitCategory
    parameters: Optional[List[ExploitParameterConfig]] = None
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__exploit"
    id: str = ""



from abc import abstractmethod

from cyst.api.environment.message import Request


class CompositeActionManager:

    @abstractmethod
    async def call_action(self, request: Request, delay: float = 0.0) -> None:
        pass

    @abstractmethod
    async def delay(self, delay: float = 0.0) -> None:
        pass



import logging

from dataclasses import dataclass, field
from enum import Enum, auto
from uuid import uuid4

from serde import serialize, coerce

from cyst.api.configuration.configuration import ConfigItem


class LogSource(Enum):
    """
    Categories of log sources.

    Possible values:
        :SYSTEM: Activities of the environment; set up, tear down, exception handling, etc.
        :MESSAGING: Logging of messages that are traversing the simulation.
        :SERVICE: Logging of active services' events.
        :MODEL: Logging of models' events.
    """
    SYSTEM = auto()
    MESSAGING = auto()
    SERVICE = auto()
    MODEL = auto()


class LogType(Enum):
    """
    Log entries structure.

    Possible values:
        :TEXT: Log entries are plain texts. One entry per line.
        :JSON: Log entries are JSON. The whole log is not JSON-compliant, because of missing top-level object.
    """
    TEXT = auto()
    JSON = auto()


@serialize(type_check=coerce)
@dataclass
class LogConfig(ConfigItem):
    """
    CYST provides what we feel are sensible defaults for logging and this configuration enables slight corrections of
    this default. The logging is built upon Python's logging framework. Each of log sources is assigned a named logger
    with the names 'system', 'messaging', 'service.', and 'model.' respectively.

    For full customization, you can disable a log by setting log_console and log_file options to False, and then you can
    use Python's logging facilities to set the system up to your liking.

    :param source: Source of the logs. While system and messaging logs will always work as expected, service and model
        logs require their authors to abide by a convention and use the logs named 'service.xxx' and 'model.xxx'.
    :type source: LogSource

    :param log_level: A level of messages to log. It is best to follow categories of Python's logging framework and use
        one of logging.(CRITICAL, ERROR, WARNING, INFO, DEBUG).
    :type log_level: int

    :param log_type: A type of logging to use. Mainly depends on whether you want a machine to process it or not.
    :type log_type: LogType

    :param log_console: Enable or disable logging to console.
    :type log_console: bool

    :param log_file: Enable or disable logging to file. If logging to file is enabled, file_path must be specified.
    :type log_file: bool

    :param file_path: A path to file that should be used for logging. If logging to file is not enabled, file is not
        opened/created.
    :type file_path: str

    """
    source: LogSource
    log_level: int
    log_type: LogType
    log_console: bool
    log_file: bool
    file_path: str = ""
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__log"
    id: str = ""


log_defaults = [
    LogConfig(
        source=LogSource.SYSTEM,
        log_level=logging.INFO,
        log_type=LogType.TEXT,
        log_console=True,
        log_file=True,
        file_path="log/cyst_system.log"
    ),
    LogConfig(
        source=LogSource.MESSAGING,
        log_level=logging.DEBUG,
        log_type=LogType.TEXT,
        log_console=True,
        log_file=True,
        file_path="log/cyst_messages.log"
    ),
    LogConfig(
        source=LogSource.SERVICE,
        log_level=logging.DEBUG,
        log_type=LogType.TEXT,
        log_console=True,
        log_file=True,
        file_path="log/cyst_service.log"
    ),
    LogConfig(
        source=LogSource.MODEL,
        log_level=logging.INFO,
        log_type=LogType.TEXT,
        log_console=True,
        log_file=False
    )
]



from dataclasses import dataclass
from typing import Tuple, Dict, Callable, List

from netaddr import IPAddress

from cyst.api.environment.configuration import EnvironmentConfiguration
from cyst.api.environment.infrastructure import EnvironmentInfrastructure
from cyst.api.environment.message import Request, Response, Status, StatusOrigin, StatusValue, StatusDetail, Message
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.policy import EnvironmentPolicy
from cyst.api.environment.platform_specification import PlatformSpecification, PlatformType
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.logic.access import AuthenticationTarget, Authorization
from cyst.api.logic.action import ActionDescription, ActionParameterType, ActionParameter, ActionType, Action
from cyst.api.logic.behavioral_model import BehavioralModel, BehavioralModelDescription
from cyst.api.logic.composite_action import CompositeActionManager
from cyst.api.network.node import Node
from cyst.api.utils.duration import Duration, msecs


@dataclass
class AuthenticationTracker:
    src_ip: str
    src_service: str
    dst_ip: str
    dst_service: str
    identity: str
    step: int
    time: int
    attempts: int


class METAModel(BehavioralModel):
    def __init__(self, configuration: EnvironmentConfiguration, resources: EnvironmentResources,
                 policy: EnvironmentPolicy, messaging: EnvironmentMessaging,
                 infrastructure: EnvironmentInfrastructure,
                 composite_action_manager: CompositeActionManager) -> None:

        self._configuration = configuration
        self._action_store = resources.action_store
        self._exploit_store = resources.exploit_store
        self._policy = policy
        self._messaging = messaging
        self._authentications: Dict[Tuple[str, str, str, str, str], AuthenticationTracker] = {}
        self._infrastructure = infrastructure
        self._cam = composite_action_manager

        self._action_store.add(ActionDescription(id="meta:inspect:node",
                                                 type=ActionType.DIRECT,
                                                 platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"), PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
                                                 description="Discovery of hosts in a network. Equivalent to ping scanning.",
                                                 parameters=[]))

        self._action_store.add(ActionDescription(id="meta:authenticate",
                                                 type=ActionType.DIRECT,
                                                 platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"), PlatformSpecification(PlatformType.REAL_TIME, "CYST")],
                                                 description="Authentication against a service.",
                                                 parameters=[ActionParameter(ActionParameterType.TOKEN, "auth_token", configuration.action.create_action_parameter_domain_any())]))

    async def action_flow(self, message: Request) -> Tuple[Duration, Response]:
        raise RuntimeError("META behavioral model does not support composite actions.")

    def action_components(self, message: Message) -> List[Action]:
        # META actions are component-less
        return []

    async def action_effect(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        if not message.action:
            raise ValueError("Action not provided")

        action_name = "_".join(message.action.fragments)
        fn: Callable[[Request, Node], Tuple[Duration, Response]] = getattr(self, "process_" + action_name, self.process_default)
        return fn(message, node)

    def process_default(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        print("Could not evaluate message. Tag in `meta` namespace unknown. " + str(message))
        return msecs(0), self._messaging.create_response(message, status=Status(StatusOrigin.SYSTEM, StatusValue.ERROR), session=message.session)

    def process_inspect_node(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        error = ""
        if message.src_ip != IPAddress("127.0.0.1"):  # Local IP has a free pass
            if not message.session:
                error = "Session not provided"
            elif message.session.end[0] not in [x.ip for x in node.interfaces]:
                error = "Session does not end in required node"

        if error:
            return msecs(0), self._messaging.create_response(message, Status(StatusOrigin.NETWORK, StatusValue.FAILURE), error, session=message.session)

        return msecs(40), self._messaging.create_response(message, Status(StatusOrigin.NODE, StatusValue.SUCCESS),
                                                          node, session=message.session, auth=message.auth)

    def process_authenticate(self, message: Request, node: Node) -> Tuple[Duration, Response]:
        # To authenticate, an actor has to go through all the phases in the authentication scheme.
        # After each factor in authentication scheme is passed, the actor is given an authorization, which enables them
        # to attempt a next step. If it is the last step, the full authorization for this authentication scheme is
        # given.

        src_node = str(message.src_ip)
        src_service = message.src_service
        dst_node = str(message.dst_ip)
        dst_service = message.dst_service

        # First of all, check if the message contains authentication token
        token_found = False
        token = None
        t = message.action.parameters.get("auth_token", None)
        if t:
            if not isinstance(t, ActionParameter):
                raise RuntimeError("Wrong object type supplied as action parameter. Type: " + str(type(t)))

            if t.type == ActionParameterType.TOKEN and t.value:
                token_found = True
                token = t.value

        if not token_found:
            return msecs(20), self._messaging.create_response(message, Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.AUTHENTICATION_NOT_PROVIDED),
                                                              "No auth token provided", session=message.session, auth=message.auth)

        #  check if node has the target service
        s = node.services.get(dst_service)
        if s is None:
            return msecs(20), self._messaging.create_response(message, Status(StatusOrigin.SERVICE, StatusValue.FAILURE,
                                                                              StatusDetail.AUTHENTICATION_NOT_PROVIDED),
                                                              "Service does not exist on this node",
                                                              session=message.session, auth=message.auth)

        #  evaluate the token
        result = self._configuration.access.evaluate_token_for_service(s, token, node, message.dst_ip)

        if result is None:
            return msecs(20), self._messaging.create_response(message, Status(StatusOrigin.SERVICE, StatusValue.FAILURE,
                                                                              StatusDetail.AUTHENTICATION_NOT_APPLICABLE),
                                                              "Token invalid for this service", session=message.session,
                                                              auth=message.auth)

        if isinstance(result, AuthenticationTarget):
            return msecs(20), self._messaging.create_response(message, Status(StatusOrigin.SERVICE, StatusValue.FAILURE,
                                                                              StatusDetail.AUTHENTICATION_NEXT),
                                                              "Continue with next factor", session=message.session,
                                                              auth=result)

        if isinstance(result, Authorization):
            return msecs(20), self._messaging.create_response(message, Status(StatusOrigin.SERVICE, StatusValue.SUCCESS),
                                                              "Authorized", session=message.session, auth=result)


def create_meta_model(configuration: EnvironmentConfiguration, resources: EnvironmentResources,
                      policy: EnvironmentPolicy, messaging: EnvironmentMessaging,
                      infrastructure: EnvironmentInfrastructure,
                      composite_action_manager: CompositeActionManager) -> BehavioralModel:
    model = METAModel(configuration, resources, policy, messaging, infrastructure, composite_action_manager)
    return model


behavioral_model_description = BehavioralModelDescription(
    namespace="meta",
    description="Interpreter for auxiliary actions needed to supplement",
    creation_fn=create_meta_model,
    platform=[PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST"), PlatformSpecification(PlatformType.REAL_TIME, "CYST")]
)


import argparse
import asyncio
import atexit
import functools
import logging
import os
import signal
import sys
import traceback

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

from heapq import heappush, heappop
from time import localtime
from typing import Tuple, List, Union, Optional, Any, Dict, Set
from threading import Condition

from cyst.api.environment.environment import Environment
from cyst.api.environment.control import EnvironmentState, EnvironmentControl
from cyst.api.environment.configuration import EnvironmentConfiguration, GeneralConfiguration, RuntimeConfiguration
from cyst.api.environment.data_model import ActionModel
from cyst.api.environment.infrastructure import EnvironmentInfrastructure
from cyst.api.environment.interpreter import ActionInterpreterDescription
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.metadata_provider import MetadataProvider
from cyst.api.environment.platform import Platform, PlatformDescription
from cyst.api.environment.platform_interface import PlatformInterface
from cyst.api.environment.platform_specification import PlatformSpecification, PlatformType
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.environment.stores import DataStoreDescription, DataStore
from cyst.api.environment.message import Message, MessageType, Request, Response
from cyst.api.logic.behavioral_model import BehavioralModelDescription, BehavioralModel
from cyst.api.network.node import Node
from cyst.api.host.service import Service
from cyst.api.configuration.configuration import ConfigItem
from cyst.api.utils.counter import Counter
from cyst.api.utils.duration import Duration

from cyst.core.environment.configuration import GeneralConfigurationImpl
from cyst.core.environment.configuration_action import ActionConfigurationImpl
from cyst.core.environment.configuration_exploit import ExploitConfigurationImpl
from cyst.core.environment.configuration_physical import PhysicalConfigurationImpl
from cyst.core.environment.environment_configuration import EnvironmentConfigurationImpl
from cyst.core.environment.environment_control import EnvironmentControlImpl
from cyst.core.environment.environment_messaging import EnvironmentMessagingImpl
from cyst.core.environment.environment_resources import EnvironmentResourcesImpl
from cyst.core.environment.infrastructure import EnvironmentInfrastructureImpl
from cyst.core.environment.stats import StatisticsImpl

from cyst.core.environment.stores import ServiceStoreImpl
from cyst.core.environment.external_resources import ExternalResourcesImpl
from cyst.core.logic.composite_action import CompositeActionManagerImpl
# from cyst.platform.environment.message import MessageImpl  # TODO: is this the correct import?


# Environment is unlike other core implementation given an underscore-prefixed name to let python complain about
# it being private if instantiated otherwise than via the create_environment()
class _Environment(Environment, PlatformInterface):

    def __init__(self, platform: Optional[Union[str, PlatformSpecification]] = None, run_id: str = "") -> None:
        self._time = 0
        self._start_time = localtime()
        self._message_queue: List[Tuple[int, int, Message]] = []
        self._executables: List[Tuple[float, int, Message, Optional[Service], Optional[Node]]] = []
        self._executed: Set[asyncio.Task] = set()
        self._pause = False
        self._terminate = False
        self._terminate_reason = ""
        self._initialized = False
        self._finish = False
        self._finish_reason = ""
        self._state = EnvironmentState.CREATED

        self._loop = asyncio.new_event_loop()
        self._loop.set_exception_handler(self.loop_exception_handler)

        self._run_id = run_id

        self._pause_on_request: List[str] = []
        self._pause_on_response: List[str] = []

        self._action_counts: Dict[str, int] = {}
        self._highest_action_count = 0
        self._highest_action_actor = ""
        self._active_actions: Dict[int, ActionModel] = {}

        # Interface implementations
        self._environment_control = EnvironmentControlImpl(self)
        self._environment_messaging = EnvironmentMessagingImpl(self)
        self._environment_resources = EnvironmentResourcesImpl(self)

        self._behavioral_models: Dict[str, BehavioralModel] = {}
        # TODO currently, there can be only on metadata provider for one namespace
        self._metadata_providers: Dict[str, MetadataProvider] = {}
        self._platforms: Dict[PlatformSpecification, PlatformDescription] = {}
        self._data_stores: Dict[str, DataStoreDescription] = {}

        self._general_configuration = GeneralConfigurationImpl(self)
        self._action_configuration = ActionConfigurationImpl()
        self._exploit_configuration = ExploitConfigurationImpl(self)
        self._physical_configuration = PhysicalConfigurationImpl(self)
        self._runtime_configuration = RuntimeConfiguration()

        self._platform = None
        self._platform_spec = None
        self._platform_notifier = Condition()

        self._configure_runtime()
        # Runtime configuration always has some value for run_id, so the run_id passed in the initializer takes a
        # precedence, if there is any
        self._run_id = self._runtime_configuration.run_id if not self._run_id else self._run_id

        self._register_metadata_providers()
        self._register_platforms()
        self._register_data_stores()

        # set a platform if it is requested
        if platform:
            platform_not_found = False
            platform_underspecified = False

            # This is rather ugly but is a price to pay for users to not need full specification
            if isinstance(platform, str):
                spec1 = PlatformSpecification(PlatformType.SIMULATED_TIME, platform)
                spec2 = PlatformSpecification(PlatformType.REAL_TIME, platform)

                spec1_flag = spec1 in self._platforms
                spec2_flag = spec2 in self._platforms

                if not spec1_flag and not spec2_flag == True:
                    platform_not_found = True
                elif spec1_flag and spec2_flag == True:
                    platform_underspecified = True
                else:
                    platform_not_found = False
                    platform = spec1 if spec1_flag else spec2
            else:
                platform_not_found = platform not in self._platforms

            if platform_not_found:
                raise RuntimeError(f"Platform {platform} is not registered into the system. Cannot continue.")

            if platform_underspecified:
                raise RuntimeError(f"Platform {platform} exists both as a simulation and realtime environment. Please, provide a full PlatformSpecification.")

            self._platform_spec = platform
        else:
            # When no specific platform is used, CYST simulation is set
            self._platform_spec = PlatformSpecification(PlatformType.SIMULATED_TIME, "CYST")

        # When platform specification is finalized, create components dependent on the platform specification and
        # components the platform depends on
        self._environment_resources = EnvironmentResourcesImpl(self, self._platform_spec)
        self._service_store = ServiceStoreImpl(self._environment_messaging, self._environment_resources, self._runtime_configuration)
        self._statistics = StatisticsImpl(self._run_id)

        if not self._runtime_configuration.data_backend in self._data_stores:
            raise ValueError(f"Required data store backend '{self._runtime_configuration.data_backend}' not installed. Cannot continue.")

        if self._runtime_configuration.data_batch_storage and self._runtime_configuration.data_backend != "memory":
            self._data_store_batch = self._data_stores[self._runtime_configuration.data_backend].creation_fn(self._run_id, self._runtime_configuration.data_backend_params)
        self._data_store = self._data_stores["memory"].creation_fn(self._run_id, {})

        self._infrastructure = EnvironmentInfrastructureImpl(self._runtime_configuration, self._data_store,
                                                             self._service_store, self._statistics)

        self._platform = self._create_platform(self._platform_spec)

        # If only there was a way to make it more sane, without needing to create a completely new interface
        self._environment_resources.init_resources(self._loop, self._platform.clock)

        # When platform is initialized, create a combined configuration for behavioral models
        self._environment_configuration = EnvironmentConfigurationImpl(self._general_configuration, self._platform.configuration,
                                                                       self._action_configuration, self._exploit_configuration,
                                                                       self._physical_configuration)

        self._cam = CompositeActionManagerImpl(self._loop, self._behavioral_models, self._environment_messaging,
                                               self._environment_resources, self._general_configuration,
                                               self._data_store, self._active_actions)

        # Services and actions depend on platform being initialized
        self._register_services()
        self._register_actions()
        self._register_metadata_providers()

        # Logs
        self._message_log = logging.getLogger("messaging")
        self._system_log = logging.getLogger("system")

        signal.signal(signal.SIGINT, self._signal_handler)
        atexit.register(self.cleanup)

    def cleanup(self):
        self._loop.close()

    def _signal_handler(self, *args):
        self._terminate = True
        self._terminate_reason = "Received a SIGINT signal to terminate."

    def loop_exception_handler(self, loop: asyncio.AbstractEventLoop, context: Dict[str, Any]) -> None:
        exception_text = str(context['exception']) + "\nCall stack: \n"
        if "future" in context:
            for frame in context["future"].get_stack():
                exception_text += f"{frame.f_code.co_filename}:{frame.f_lineno} :: {frame.f_code.co_qualname}\n"
        else:
            trace = traceback.StackSummary.extract(traceback.walk_tb(context["exception"].__traceback__))
            for frame in trace:
                exception_text += f"{frame.filename}:{frame.line} :: {frame.name}\n"

        error = f"Unhandled exception in event loop. Exception: {exception_text}."
        print(error)
        self._terminate = True
        self._terminate_reason = error

    def __getstate__(self) -> dict:
        return {
            # Simple values
            "_time": self._time,
            "_start_time": self._start_time,
            "_pause": self._pause,
            "_terminate": self._terminate,
            "_initialized": self._initialized,
            "_state": self._state,
            "_run_id": self._run_id,

            # Arrays
            "_pause_on_response": self._pause_on_response,
            "_pause_on_request": self._pause_on_request,

            # Simple objects
            "_runtime_configuration": self._runtime_configuration,

            # Complex beasts
            "_service_store": self._service_store,
            "_environment_resources": self._environment_resources,
            "_metadata_providers": self._metadata_providers,
            "_general_configuration": self._general_configuration

            # Ignored members
            # Policy - is reinitialized, no need to serialize
            # DataStore - stays the same across serializations
            # Log - stays the same across serializations
            # All interface implementations excluding the general configuration and environment resources
        }

    def __setstate__(self, state: dict) -> None:
        self._time = state["_time"]
        self._start_time = state["_start_time"]
        self._pause = state["_pause"]
        self._terminate = state["_terminate"]
        self._initialized = state["_initialized"]
        self._state = state["_state"]
        self._run_id = state["_run_id"]

        self._pause_on_response = state["_pause_on_response"]
        self._pause_on_request = state["_pause_on_request"]

        self._runtime_configuration = state["_runtime_configuration"]

        self._service_store = state["_service_store"]
        self._environment_resources = state["_environment_resources"]
        self._metadata_providers = state["_metadata_providers"]
        self._general_configuration = state["_general_configuration"]

        self._environment_control = EnvironmentControlImpl(self)
        self._environment_messaging = EnvironmentMessagingImpl(self)

    # Replace the environment with the state of another environment. This is used for deserialization. It is explicit to
    # avoid replacing of ephemeral stuff, such as data store connections or whatnot
    def _replace(self, env: "_Environment"):
        self._time = env._time
        self._start_time = env._start_time
        self._pause = env._pause
        self._terminate = env._terminate
        self._initialized = env._initialized
        self._state = env._state
        self._run_id = env._run_id

        self._pause_on_response = env._pause_on_response
        self._pause_on_request = env._pause_on_request

        self._runtime_configuration = env._runtime_configuration

        self._service_store = env._service_store
        self._environment_resources = env._environment_resources
        self._metadata_providers = env._metadata_providers
        self._general_configuration = env._general_configuration

    # Runtime parameters can be passed via command-line, configuration file, or through environment variables
    # In case of multiple definitions of one parameter, the order is, from the most important to least:
    #                                                            command line, configuration file, environment variables
    def _configure_runtime(self) -> None:
        # Environment
        data_backend = os.environ.get('CYST_DATA_BACKEND')
        data_backend_params: Dict[str, str] = dict()
        if data_backend:
            data_backend_params_serialized = os.environ.get('CYST_DATA_BACKEND_PARAMS')
            # we expect parameters to be given in the form "param1_name","param1_value","param2_name","param2_value",...
            if data_backend_params_serialized:
                data_backend_params_list = [x.strip() for x in data_backend_params_serialized.split(",")]
                data_backend_params = dict([(data_backend_params_list[i], data_backend_params_list[i+1]) for i in range(0, len(data_backend_params_list), 2)])

        data_batch_storage = False
        if "CYST_DATA_BATCH_STORAGE" in os.environ:
            data_batch_storage = True

        run_id = os.environ.get('CYST_RUN_ID')
        config_id = os.environ.get('CYST_CONFIG_ID')

        max_running_time_s = os.environ.get("CYST_MAX_RUNNING_TIME")
        if max_running_time_s:
            max_running_time = float(max_running_time_s)
        else:
            max_running_time = 0.0

        run_id_log_suffix = False
        if "CYST_RUN_ID_LOG_SUFFIX" in os.environ:
            run_id_log_suffix = True

        max_action_count_s = os.environ.get("CYST_MAX_ACTION_COUNT")
        if max_action_count_s:
            max_action_count = int(max_action_count_s)
        else:
            max_action_count = 0

        config_filename = os.environ.get('CYST_CONFIG_FILENAME')

        # All the unknown, CYST-related params
        # TODO: Where to document this?
        for k, v in os.environ.items():
            if k.startswith("CYST_") and k not in ["CYST_DATA_BACKEND", "CYST_DATA_BACKEND_PARAMS", "CYST_RUN_ID",
                                                   "CYST_CONFIG_ID", "CYST_MAX_RUNNING_TIME", "CYST_RUN_ID_LOG_SUFFIX",
                                                   "CYST_MAX_ACTION_COUNT", "CYST_CONFIG_FILENAME", "CYST_DATA_BATCH_STORAGE"]:
                name = k[5:].lower()
                self._runtime_configuration.other_params[name] = v

        # Command line (only parsing)
        cmdline_parser = argparse.ArgumentParser(description="CYST runtime configuration")

        cmdline_parser.add_argument("-c", "--config_file", type=str,
                                    help="Path to a file storing the configuration. Commandline overrides the items in configuration file.")
        cmdline_parser.add_argument("-b", "--data_backend", type=str,
                                    help="The type of a backend to use. Currently supported are: memory, sqlite")
        cmdline_parser.add_argument("-p", "--data_backend_parameter", action="append", nargs=2, type=str,
                                    metavar=('NAME', 'VALUE'), help="Parameters to be passed to data backend.")
        cmdline_parser.add_argument("-g", "--data_batch_storage", type=bool,
                                    help="Store data in memory and move them to other backend on termination.")
        cmdline_parser.add_argument("-r", "--run_id", type=str,
                                    help="A unique identifier of a simulation run. If not specified, a UUID will be generated instead.")
        cmdline_parser.add_argument("-i", "--config_id", type=str,
                                    help="A unique identifier of simulation run configuration, which can be obtained from the data store.")
        cmdline_parser.add_argument("-t", "--max_running_time", type=float,
                                    help="An upper limit on an execution time of a run. A platform time is considered, not the real time.")
        cmdline_parser.add_argument("-a", "--max_action_count", type=int,
                                    help="An upper limit on the number of executed actions by any actor. When this count is reached, the run terminates.")
        cmdline_parser.add_argument("-s", "--run_id_log_suffix", type=bool,
                                    help="Set to true if you want log file names to have the run id as a suffix.")
        cmdline_parser.add_argument("-o", "--other_param", action="append", nargs=2, type=str, metavar=('NAME', 'VALUE'),
                                    help="Other parameters that are passed to CYST components, agents, etc.")

        args, _ = cmdline_parser.parse_known_args()
        if args.config_file:
            config_filename = args.config_file

        # --------------------------------------------------------------------------------------------------------------
        # Config file TODO
        if config_filename:
            pass
        # --------------------------------------------------------------------------------------------------------------

        # Command line argument override
        if args.data_backend:
            data_backend = args.data_backend

        if args.data_backend_parameter:
            # Convert from list of lists into a list of tuples
            data_backend_params = dict(tuple(x) for x in args.data_backend_parameter) #MYPY: typehinting lambda not really possible this way, better to ignore?

        if args.data_batch_storage:
            data_batch_storage = args.data_batch_storage

        if args.run_id:
            run_id = args.run_id

        if args.config_id:
            config_id = args.config_id

        if args.max_running_time:
            max_running_time = args.max_running_time

        if args.max_action_count:
            max_action_count = args.max_action_count

        if args.run_id_log_suffix:
            run_id_log_suffix = args.run_id_log_suffix

        if args.other_param:
            for x in args.other_param:
                name = x[0].lower()
                self._runtime_configuration.other_params[name] = x[1]

        # --------------------------------------------------------------------------------------------------------------
        if data_backend:  # Fuck, I miss oneliners
            self._runtime_configuration.data_backend = data_backend
        if data_backend_params:
            self._runtime_configuration.data_backend_params = data_backend_params
        self._runtime_configuration.data_batch_storage = data_batch_storage
        if config_filename:
            self._runtime_configuration.config_filename = config_filename
        if run_id:
            self._runtime_configuration.run_id = run_id
        if config_id:
            self._runtime_configuration.config_id = config_id
        if max_running_time:
            self._runtime_configuration.max_running_time = max_running_time
        if max_action_count:
            self._runtime_configuration.max_action_count = max_action_count
        self._runtime_configuration.run_id_log_suffix = run_id_log_suffix

    def configure(self, *config_item: ConfigItem, parameters: dict[str, Any] | None = None) -> Environment:
        # Preprocess all configuration items for easier platform management
        self._general_configuration.preprocess(parameters, *config_item)
        # Configure general stuff
        self._general_configuration.configure()
        # Process the rest in platform
        self._platform.configure(*self._general_configuration.get_configuration())

        return self

    # ------------------------------------------------------------------------------------------------------------------
    # Environment interfaces
    # ------------------------------------------------------------------------------------------------------------------
    @property
    def general(self) -> GeneralConfiguration:
        return self._general_configuration

    @property
    def configuration(self) -> EnvironmentConfiguration:
        return self._platform.configuration

    @property
    def control(self) -> EnvironmentControl:
        return self._environment_control

    @property
    def messaging(self) -> EnvironmentMessaging:
        return self._environment_messaging

    @property
    def platform_interface(self) -> PlatformInterface:
        return self

    @property
    def platform(self) -> Platform:
        return self._platform

    @property
    def resources(self) -> EnvironmentResources:
        return self._environment_resources

    @property
    def infrastructure(self) -> EnvironmentInfrastructure:
        return self._infrastructure

    # ------------------------------------------------------------------------------------------------------------------
    # An interface between the environment and a platform
    def execute_task(self, task: Message, service: Optional[Service] = None, node: Optional[Node] = None, delay: float = 0.0) -> Tuple[bool, int]:
        heappush(self._executables, (self._platform.clock.current_time() + delay, Counter().get("msg"), task, service, node))

        return True, 0

    def process_response(self, response: Response, delay: float = 0.0) -> Tuple[bool, int]:
        self.messaging.send_message(response, delay)
        return True, 0

    # ------------------------------------------------------------------------------------------------------------------
    # Internal functions
    # TODO: Bloody names!
    def _process_finalized_task(self, task: asyncio.Task) -> None:
        try:
            delay, response = task.result()
        except Exception:  # Too broad... yeah, whatever
            call_stack = ""
            for frame in task.get_stack():
                call_stack += f"{frame.f_code.co_filename}:{frame.f_lineno} :: {frame.f_code.co_qualname}\n"
            self._system_log.log(logging.ERROR, f"There was an exception when running a task: {repr(task.exception())}.\nCall stack:\n{call_stack}")
            self._terminate = True
            self._terminate_reason = f"There was an exception when running a task: {repr(task.exception())}."
            return

        # TODO: Leaving this here, until Duration is everywhere
        if isinstance(delay, Duration):
            delay = delay.to_float()
        self.process_response(response, delay)
        self._executed.remove(task)

    def _finalize_process_message(self, message: Message, caller_id: str, task: asyncio.Task) -> None:
        #  TODO: Do we need to process the result?
        # success, delay = task.result()
        self._executed.remove(task)

        if message.type == MessageType.RESPONSE:
            if message.id in self._active_actions:
                active_action = self._active_actions[message.id]
                active_action.set_response(message)
                self._data_store.add_action(active_action)
                del self._active_actions[message.id]

            if caller_id in self._pause_on_response:
                self._pause = True

    async def _process_async(self) -> None:
        # Message sending tasks are delegated to platforms
        # Execution of behavioral models, composite actions and external resources are handled by the environment
        current_time = self._platform.clock.current_time()
        time_jump = 0

        # Set the flag to finish if we exceed the max running time.
        if self._runtime_configuration.max_running_time > 0.0 and not self._finish:
            if current_time > self._runtime_configuration.max_running_time:
                self._finish = True
                self._finish_reason = f"Exceeded the time limit of {self._runtime_configuration.max_running_time} virtual seconds "
                self._system_log.info(f"Terminating run because we ran over the time limit of {self._runtime_configuration.max_running_time} virtual seconds.")
                return

        # Set the flag to finish if we exceed the max action count from one source.
        if self._runtime_configuration.max_action_count > 0 and not self._finish:
            if self._highest_action_count > self._runtime_configuration.max_action_count:
                self._finish = True
                self._finish_reason = f"The actor '{self._highest_action_actor}' crossed the action limit of {self._runtime_configuration.max_action_count}."
                self._system_log.info(f"Terminating run because the actor '{self._highest_action_actor}' crossed the action limit of {self._runtime_configuration.max_action_count}.")
                return

        have_something_to_do = bool(self._executables) or bool(self._executed)

        # --------------------------------------------------------------------------------------------------------------
        # Process the resources if there are any
        ext = ExternalResourcesImpl.cast_from(self._environment_resources.external)
        if self._platform_spec.type == PlatformType.SIMULATED_TIME:
            ext.collect_at(current_time)
            # Suggest a time jump if there are resources waiting to be processed. Otherwise, it would just be set to 0.
            time_jump = ext.pending()[1]
        else:
            # No time jump is suggested, because time runs its own course
            ext.collect_immediately()

        # --------------------------------------------------------------------------------------------------------------
        # We let the composite action manager start all the tasks
        # This is almost no-op if no requests are in a queue for it. And if there are, they will just be processed and
        # converted to normal messages down the line.
        # Note on that |= ... process returns bool if there is some processing being done
        cam_queues_left, composite_actions_resolving, composite_actions_processing = await self._cam.process()
        have_something_to_do |= cam_queues_left

        # --------------------------------------------------------------------------------------------------------------
        # Get the required time delta
        if self._executables:
            next_time = self._executables[0][0]
            delta = next_time - current_time
            if time_jump == 0 or delta < time_jump:
                time_jump = delta

        # --------------------------------------------------------------------------------------------------------------
        # If there are still some resources that are being worked on (i.e., process after collection) than we forbid
        # the time jump
        if ext.collecting() or composite_actions_resolving:
            time_jump = 0
            have_something_to_do = True

        # --------------------------------------------------------------------------------------------------------------
        # If there is a time to jump, instruct the platform to do so
        platform_has_something_to_do = False
        if not have_something_to_do or time_jump > 0:
            platform_has_something_to_do = await self._platform.process(time_jump)
            # Return to have the process started anew
            if platform_has_something_to_do:
                return

        # Nothing pending in queues
        if not (have_something_to_do or platform_has_something_to_do or composite_actions_resolving or composite_actions_processing or ext.active()):
            # This is here to make it explicit - we do not want to finish if we are still in INIT, as this means we are
            # still executing run() methods of active services.
            if self._state != EnvironmentState.INIT:
                self._finish = True
                self._finish_reason = "No activity pending in the run."
                return

        # --------------------------------------------------------------------------------------------------------------
        # Task gathering
        tasks_to_execute = []

        # Tasks scheduled for execution
        if self._executables:
            next_time = self._executables[0][0]
            while next_time <= current_time:
                task = heappop(self._executables)
                tasks_to_execute.append((task[2], task[3], task[4]))
                if self._executables:
                    next_time = self._executables[0][0]
                else:
                    break

        for task in tasks_to_execute:
            message = task[0]
            service = task[1]
            node = task[2]

            # If the task is a part of composite processing, then pass it to cam
            if self._cam.is_composite(message.id) and message.type == MessageType.RESPONSE:
                self._cam.incoming_message(message)

            # If an active service is provided, we are calling its process_message method. Otherwise, behavioral model
            # is invoked.
            elif service and service.active_service:
                # Extract and clear platform-specific information
                caller_id = ""
                if message.type == MessageType.RESPONSE:
                    caller_id = message.platform_specific["caller_id"] if "caller_id" in message.platform_specific else ""
                    message.platform_specific.clear()

                # service.active_service.process_message(message)
                t = self._loop.create_task(service.active_service.process_message(message))
                self._executed.add(t)
                t.add_done_callback(functools.partial(self._finalize_process_message, message, caller_id))
            else:
                request = message.cast_to(Request)
                namespace = request.action.namespace
                t = self._loop.create_task(self._behavioral_models[namespace].action_effect(request, node))
                self._executed.add(t)
                t.add_done_callback(self._process_finalized_task)

    def _register_services(self) -> None:

        # First, check entry points registered via the importlib mechanism
        plugin_services = entry_points(group="cyst.services")
        for s in plugin_services:
            service_description = s.load()

            if self._service_store.get_service(service_description.name):
                print("Service with name {} already registered, skipping...".format(service_description.name))
            else:
                self._service_store.add_service(service_description)

        # Explicit addition of built-in active services
        # self._service_store.add_service(firewall_service_description)

    def _register_actions(self) -> None:

        plugin_models = entry_points(group="cyst.models")
        for s in plugin_models:
            model_description = s.load()

            if not isinstance(model_description, BehavioralModelDescription):
                if isinstance(model_description, ActionInterpreterDescription):
                    print(f"The model of namespace '{model_description.namespace}' uses the old API specification. "
                          f"From version 0.6.0 only BehavioralModelDescription is supported. This model will be ignored.")
                    continue
                raise RuntimeError(f"Model of unsupported type [{type(model_description)}] intercepted. Please, fix the installation.")

            model_platform = model_description.platform
            if not isinstance(model_platform, list):
                model_platform = [model_platform]

            # Skip behavioral models not supported for this platform
            if self._platform_spec not in model_platform:
                continue

            if model_description.namespace in self._behavioral_models:
                print("Behavioral model with namespace {} already registered, skipping it ...".format(
                    model_description.namespace))
            else:
                model = model_description.creation_fn(self._environment_configuration, self._environment_resources, None,
                                                      self._environment_messaging, self._infrastructure, self._cam)
                self._behavioral_models[model_description.namespace] = model

    def _register_metadata_providers(self) -> None:

        plugin_providers = entry_points(group="cyst.metadata_providers")
        for s in plugin_providers:
            provider_description = s.load()

            if provider_description.namespace in self._metadata_providers:
                print("Metadata provider with namespace {} already registered, skipping ...".format(
                    provider_description.namespace))
            else:
                provider = provider_description.creation_fn()
                self._metadata_providers[provider_description.namespace] = provider

    def _register_platforms(self) -> None:

        plugin_providers = entry_points(group="cyst.platforms")
        for s in plugin_providers:
            platform_description = s.load()

            if platform_description.specification in self._platforms:
                print("Platform with specification {} already registered, skipping ...".format(
                    str(platform_description.specification)))
            else:
                self._platforms[platform_description.specification] = platform_description

    def _register_data_stores(self) -> None:

        plugin_providers = entry_points(group="cyst.data_stores")
        for s in plugin_providers:
            data_store_description = s.load()

            if data_store_description.backend in self._data_stores:
                print(f"Data store with backend {data_store_description.backend} already registered, skipping ...")
            else:
                self._data_stores[data_store_description.backend] = data_store_description

    def _create_platform(self, specification: PlatformSpecification) -> Platform:
        if specification not in self._platforms:
            raise RuntimeError(f"Attempting to create a platform that is not registered: {specification}")

        return self._platforms[specification].creation_fn(self.platform_interface, self._general_configuration,
                                                          self.resources, self._action_configuration,
                                                          self._exploit_configuration, self._physical_configuration,
                                                          self._infrastructure)


def create_environment(platform: Optional[Union[str, PlatformSpecification]] = None, run_id: str = "") -> Environment:
    e = _Environment(platform, run_id)
    return e



from cyst.api.configuration.configuration import ConfigItem, ConfigParameter, ConfigParameterSingle, ConfigParameterGroup, ConfigParameterValueType, ConfigParameterGroupType, ConfigParameterGroupEntry, ConfigParametrization
from cyst.api.configuration.host.service import ActiveServiceConfig, PassiveServiceConfig, ServiceParameter
from cyst.api.configuration.logic.access import AccessLevel, AuthorizationConfig, AccessSchemeConfig, AuthenticationProviderConfig, AuthorizationDomainConfig, AuthorizationDomainType, FederatedAuthorizationConfig
from cyst.api.configuration.logic.data import DataConfig
from cyst.api.configuration.logic.exploit import ExploitLocality, ExploitCategory, ExploitParameterType, \
     VulnerableServiceConfig, ExploitParameterConfig, ExploitConfig
from cyst.api.configuration.network.network import NetworkConfig
from cyst.api.configuration.network.node import NodeConfig
from cyst.api.configuration.network.elements import IPAddress, IPNetwork, ConnectionConfig, PortConfig, InterfaceConfig, RouteConfig, SessionConfig
from cyst.api.configuration.network.router import RouterConfig
from cyst.api.configuration.network.firewall import FirewallRule, FirewallChainType, FirewallChainConfig, FirewallPolicy, FirewallConfig

__all__ = ["ActiveServiceConfig", "PassiveServiceConfig", "AccessLevel", "AuthorizationConfig", "DataConfig", "ExploitLocality",
           "ExploitCategory", "ExploitParameterType", "VulnerableServiceConfig", "ExploitParameterConfig", "ExploitConfig",
           "NetworkConfig", "NodeConfig", "IPAddress", "IPNetwork", "ConnectionConfig", "PortConfig", "InterfaceConfig", "RouterConfig",
           "FirewallRule", "FirewallChainType", "FirewallChainConfig", "FirewallPolicy", "FirewallConfig", "ServiceParameter",
           "AccessSchemeConfig", "AuthenticationProviderConfig", "AuthorizationDomainConfig", "AuthorizationDomainType",
           "FederatedAuthorizationConfig", "RouteConfig", "ConfigItem", "ConfigParameter", "ConfigParametrization", "ConfigParameterGroupEntry",
           "ConfigParameterGroupType", "ConfigParameterSingle", "ConfigParameterGroup", "ConfigParameterValueType", "SessionConfig"]



from abc import ABC, abstractmethod
from dataclasses import dataclass
from deprecated.sphinx import versionchanged
from typing import Callable

from cyst.api.environment.configuration import ActionConfiguration
from cyst.api.environment.stores import ActionStore
from cyst.api.environment.message import Message
from cyst.api.logic.action import Action
from cyst.api.logic.metadata import Metadata


class MetadataProvider(ABC):
    """
    Metadata providers supply messages with additional information that are meant to mimic observations that can be
    done in the real life. Two typical examples of metadata can be 1) flow information, 2) results of traffic analyses.
    """
    @versionchanged(version="0.6.0", reason="Metadata are collected for a whole message at the time of sending and are no longer solely dependent on Action.")
    @abstractmethod
    def get_metadata(self, action: Action, message: Message) -> Metadata:
        """
        This function should return metadata that correspond to the given message.

        :param action: The action for which the metadata should be provided. Given that a message can carry multiple
            actions, a metadata provider must adhere to this parameter and not try to extract it from the message
            itself.
        :type action: Action

        :param message: An instance of the message. The message should be used to provide additional information
            necessary to supply correct metadata, such as message type, message status, etc.
        :type message: Message

        :return: A metadata associated with the message.
        """
        pass


@dataclass
class MetadataProviderDescription:
    """ A description of a metadata provider which should be registered into the system.

    :param namespace: A namespace of actions that this provider works with.
    :type namespace: str

    :param description: A short description of the provider.
    :type description: str

    :param creation_fn: A factory function that creates instances of the metadata provider.
    :type creation_fn: Callable[[ActionStore, ActionConfiguration], MetadataProvider]
    """
    namespace: str
    description: str
    creation_fn: Callable[[], MetadataProvider]



from dataclasses import dataclass, field

from netaddr import IPAddress, IPNetwork
from typing import Union, Optional, List
from uuid import uuid4
from serde import serialize, coerce
from serde.compat import typename

from cyst.api.configuration.configuration import ConfigItem


@serialize(type_check=coerce)
@dataclass
class PortConfig(ConfigItem):
    """ Configuration of a network port.

    A network port represents an abstraction of an ethernet port, with a given IP address and a given network. A network
    port does not support a default routing through a gateway and so it is used mostly for routers, which maintain
    their own routing tables based on the port indexes.

    :param ip: The assigned IP address of the port.
    :type ip: IPAddress

    :param net: The assigned network of the port. If used only for inter-router communication, ip/32 or ip/128
        can be used.
    :type net: IPNetwork

    :param index: The index of the port. The index is used for unique addressing of a port within a node, especially
        for correctly setting routing tables. If left at the default value, it is assigned the next free index.
    :type index: int
    """
    ip: IPAddress = field(metadata={
        'serde_serializer': lambda x: {"cls_type": typename(type(x)), "value": str(x)},
    })
    net: IPNetwork = field(metadata={
        'serde_serializer': lambda x: {"cls_type": typename(type(x)), "value": str(x)},
    })
    index: int = field(default=-1)
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__port"
    id: str = ""


@serialize(type_check=coerce)
@dataclass
class InterfaceConfig(ConfigItem):
    """ Configuration of a network interface.

    A network interface represents an abstraction of an ethernet port, with a given IP address and a given network.
    A network interface automatically calculates the gateway and therefore enables a seamless networking.

    :param ip: The assigned IP address of the interface.
    :type ip: IPAddress

    :param net: The assigned network of the interface.
    :type net: IPNetwork

    :param index: The index of the interface. The index is used for unique addressing of an interface within a node. If
        left at the default value, it is assigned the next free index.
    :type index: int
    """
    ip: IPAddress = field(metadata={
        'serde_serializer': lambda x: {"cls_type": typename(type(x)), "value": str(x)}
    })
    net: IPNetwork = field(metadata={
        'serde_serializer': lambda x: {"cls_type": typename(type(x)), "value": str(x)}
    })
    index: int = field(default=-1)
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__interface"
    id: str = ""


@serialize(type_check=coerce)
@dataclass
class ConnectionConfig(ConfigItem):
    """ Configuration of a network connection.

    Represents a connection between two network ports/interfaces. A connection will in future support setting of
    connection properties, such as delay or packet drops. While the supporting infrastructure is partially present
    in the code now, it is not propagated into the configuration a so, each connection has a unit speed (in terms of
    the simulation time) with zero drops.

    :param src_ref: Connection source configuration or its ref.
    :type src_ref: ConfigItem | str

    :param src_port: The index of a source port/interface.
    :type src_port: int

    :param dst_ref: Connection destination configuration or its ref.
    :type dst_ref: ConfigItem | str

    :param dst_port: The index of a destination port/interface.
    :type dst_port: int
    """
    src_ref: Union[ConfigItem, str]
    src_port: int
    dst_ref: Union[ConfigItem, str]
    dst_port: int
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__connection"
    id: str = ""

    def __post_init__(self):
        if isinstance(self.src_ref, ConfigItem):
            self.src_ref = self.src_ref.ref

        if isinstance(self.dst_ref, ConfigItem):
            self.dst_ref = self.dst_ref.ref

        if not self.src_ref or not self.dst_ref:
            raise RuntimeError(f"Connection configuration can't have an empty source or destination. Source: {self.src_ref}, Destination: {self.dst_ref}.")


@serialize(type_check=coerce)
@dataclass
class RouteConfig(ConfigItem):
    """ Configuration of a network route.

    A route specifies which port should the traffic to specific network be routed through. Many routes can be specified
    for a node. If there is an overlap in network specification, than the resulting port is selected according to the
    route metrics (i.e., the lower the metric the higher the chance to be selected as the route). In case of a metric
    equality, the most specific network is selected.

    :param network: A network this route is related to.
    :type network: IPNetwork

    :param port: A port/interface index, where to route traffic to the particular network.
    :type port: int

    :param metric: A route metric used for deciding which route to use in case of network overlap.
    :type metric: int
    """
    network: IPNetwork
    port: int
    metric: int = field(default=100)
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__route"
    id: str = ""


@serialize(type_check=coerce)
@dataclass
class SessionConfig(ConfigItem):
    """ Configuration of a session.

    A session represents a virtual connection between two endpoint services that is ignoring the routing limitations
    imposed by routers and firewalls en route.

    :param src_service: A service that the session originates from.
    :type src_service: str

    :param dst_service: A service that the session terminates at.
    :type dst_service: str

    :param waypoints: A list of node IDs through which the session is established.
    :type waypoints: List[str]

    :param reverse: The direction of construction of the session. If reverse is set, then session is originating from
        the destination service. This can matter, because the session creation should honor the router configuration and
        the reverse shell can be the only one that can be established. But, for example, the CYST simulation engine
        tries to construct the session in both ways to see if anything works, so this parameter may be inconsequential.
    :type reverse: bool
    """
    src_service: str
    dst_service: str
    # TODO: Session configuration does not enable chaining through parent-child relation. This may be wanted in some cases.
    waypoints: List[str]
    reverse: bool = False
    ref: str = field(default_factory=lambda: str(uuid4()))
    name: str = "__session"
    id: str = ""



from __future__ import annotations

import uuid

from typing import TYPE_CHECKING, List, Optional, Union, Tuple

from netaddr import IPAddress

from cyst.api.environment.configuration import AccessConfiguration
from cyst.api.host.service import Service, PassiveService
from cyst.api.logic.access import AuthenticationToken, Authorization, AuthenticationTarget, AccessScheme, \
    AuthenticationProvider, AccessLevel, AuthenticationTokenType, AuthenticationTokenSecurity, \
    AuthenticationProviderType
from cyst.api.network.node import Node

from cyst.platform.host.service import PassiveServiceImpl
from cyst.platform.logic.access import AuthenticationTokenImpl, AuthenticationProviderImpl, AccessSchemeImpl, \
    AuthorizationImpl, AuthenticationTargetImpl
from cyst.platform.network.node import NodeImpl

if TYPE_CHECKING:
    from cyst.platform.main import CYSTPlatform


class AccessConfigurationImpl(AccessConfiguration):
    def __init__(self, platform: CYSTPlatform):
        self._platform = platform

    def create_authentication_provider(self, provider_type: AuthenticationProviderType,
                                       token_type: AuthenticationTokenType, security: AuthenticationTokenSecurity,
                                       ip: Optional[IPAddress], timeout: int, id: str = "") -> AuthenticationProvider:
        return _create_authentication_provider(self._platform, provider_type, token_type, security, ip, timeout, id)

    def create_authentication_token(self, type: AuthenticationTokenType, security: AuthenticationTokenSecurity,
                                    identity: str, is_local: bool) -> AuthenticationToken:
        return AuthenticationTokenImpl(type, security, identity, is_local)._set_content(uuid.uuid4())
        # content setting is temporary until encrypted/hashed data is implemented

    def register_authentication_token(self, provider: AuthenticationProvider, token: AuthenticationToken) -> bool:
        if isinstance(provider, AuthenticationProviderImpl):
            provider.add_token(token)
            return True

        return False

    def create_and_register_authentication_token(self, provider: AuthenticationProvider, identity: str) -> Optional[
        AuthenticationToken]:
        if isinstance(provider, AuthenticationProviderImpl):
            token = self.create_authentication_token(provider.token_type, provider.security, identity,
                                                     True if provider.type == AuthenticationProviderType.LOCAL else False)
            self.register_authentication_token(provider, token)
            return token

        return None

    def create_authorization(self, identity: str, access_level: AccessLevel, id: str, nodes: Optional[List[str]] = None,
                             services: Optional[List[str]] = None) -> Authorization:
        return _create_authorization(self._platform, identity, access_level, id, nodes, services)

    def create_access_scheme(self, id: str = "") -> AccessScheme:
        return _create_access_scheme(self._platform, id)

    def add_provider_to_scheme(self, provider: AuthenticationProvider, scheme: AccessScheme) -> None:
        if isinstance(scheme, AccessSchemeImpl):
            scheme.add_provider(provider)
        else:
            raise RuntimeError("Attempted to provide a malformed object with AccessScheme interface")

    def add_authorization_to_scheme(self, auth: Authorization, scheme: AccessScheme) -> None:
        if isinstance(scheme, AccessSchemeImpl):
            scheme.add_authorization(auth)
            scheme.add_identity(auth.identity)
        else:
            raise RuntimeError("Attempted to provide a malformed object with AccessScheme interface")

    def evaluate_token_for_service(self, service: Service, token: AuthenticationToken, node: Node,
                                   fallback_ip: Optional[IPAddress]) -> Optional[
        Union[Authorization, AuthenticationTarget]]:
        # check if node has the service is in interpreter
        if isinstance(service, PassiveServiceImpl):
            for scheme in service.access_schemes:
                result = _assess_token(self._platform, scheme, token)
                if isinstance(result, Authorization):
                    return _user_auth_create(self._platform, result, service, node)
                if isinstance(result, AuthenticationTargetImpl):
                    if result.address is None:
                        result.address = fallback_ip
                    return result

        return None

    def unregister_authentication_token(self, token_identity: str, provider: AuthenticationProvider) -> None:
        AuthenticationProviderImpl.cast_from(provider).remove_token_by_identity(token_identity)

    def disable_authentication_token(self, provider: AuthenticationProvider, token: AuthenticationToken, time: int) -> None:
        AuthenticationProviderImpl.cast_from(provider).disable_token(token, time)

    def enable_authentication_token(self, provider: AuthenticationProvider, token: AuthenticationToken) -> None:
        AuthenticationProviderImpl.cast_from(provider).enable_token(token)

    def remove_authorization_from_scheme(self, auth: Authorization, scheme: AccessScheme) -> None:
        AccessSchemeImpl.cast_from(scheme).remove_authorization(auth)

    def modify_existing_access(self, service: Service, identity: str, access_level: AccessLevel) -> bool:
        if isinstance(service, PassiveService):
            ok, _ = self._access_service(service, identity, access_level, allow_modify=True)
            return ok
        return False

    def create_service_access(self, service: Service, identity: str, access_level: AccessLevel,
                              tokens: List[AuthenticationToken] = None) -> Optional[List[AuthenticationToken]]:
        if isinstance(service, PassiveService):
            ok, tokens = self._access_service(service, identity, access_level, tokens)
            if ok:
                return tokens
        return None

    @staticmethod
    def _access_service(service: PassiveService, identity: str, access_level: AccessLevel,
                        tokens: List[AuthenticationToken] = None, allow_modify: bool = False) -> Tuple[bool, List[AuthenticationToken]]:
        for scheme in PassiveServiceImpl.cast_from(service)._access_schemes:
            actual_scheme = AccessSchemeImpl.cast_from(scheme)
            if not actual_scheme.is_local():
                # TODO: handle non-local providers
                continue

            if allow_modify and identity in actual_scheme.identities:
                actual_scheme.modify_access(identity, access_level)
                return True, []

            elif not allow_modify and identity not in actual_scheme.identities:
                if tokens and not actual_scheme.match_tokens(tokens):
                    continue
                return True, actual_scheme.create_access(identity, access_level, tokens)

        return False, []


# ------------------------------------------------------------------------------------------------------------------
# Access configuration
def _create_authentication_provider(self: CYSTPlatform, provider_type: AuthenticationProviderType,
                                   token_type: AuthenticationTokenType, security: AuthenticationTokenSecurity,
                                   ip: Optional[IPAddress], timeout: int, id: str = "") -> AuthenticationProvider:
    if not id:
        id = str(uuid.uuid4())
    a = AuthenticationProviderImpl(provider_type, token_type, security, ip, timeout, id)
    self._general_configuration.add_object(a.id, a)
    return a


def _create_authorization(self: CYSTPlatform, identity: str, access_level: AccessLevel, id: str, nodes: Optional[List[str]] = None,
                         services: Optional[List[str]] = None) -> Authorization:
    if not id:
        id = str(uuid.uuid4())

    a = AuthorizationImpl(
        identity=identity,
        access_level=access_level,
        id=id,
        nodes=nodes,
        services=services
    )

    self._general_configuration.add_object(id, a)
    return a


def _create_access_scheme(self: CYSTPlatform, id: str = "") -> AccessScheme:
    if not id:
        id = str(uuid.uuid4())
    scheme = AccessSchemeImpl(id)
    self._general_configuration.add_object(scheme.id, scheme)
    return scheme


def _assess_token(self: CYSTPlatform, scheme: AccessScheme, token: AuthenticationToken) \
        -> Optional[Union[Authorization, AuthenticationTarget]]:

    for i in range(0, len(scheme.factors)):
        if scheme.factors[i][0].token_is_registered(token):
            if i == len(scheme.factors) - 1:
                return next(filter(lambda auth: auth.identity == token.identity, scheme.authorizations), None)
            else:
                return scheme.factors[i + 1][0].target
    return None


def _user_auth_create(self: CYSTPlatform, authorization: Authorization, service: Service, node: Node):
    if isinstance(authorization, AuthorizationImpl):
        if (authorization.nodes == ['*'] or NodeImpl.cast_from(node).id in authorization.nodes) and \
                (authorization.services == ['*'] or service.name in authorization.services):

            ret_auth = AuthorizationImpl(
                identity=authorization.identity,
                nodes=[NodeImpl.cast_from(node).id],
                services=[service.name],
                access_level=authorization.access_level,
                id=str(uuid.uuid4())
            )

            if isinstance(service, PassiveServiceImpl):
                service.add_active_authorization(ret_auth)  # TODO: check if this can go to public/private auths
            return ret_auth
    return None



from abc import ABC, abstractmethod
from dataclasses import dataclass
from deprecated.sphinx import versionchanged
from enum import IntEnum
from semver import VersionInfo
from typing import Set, Tuple, Callable, Dict, Any, Optional

from cyst.api.environment.message import Message
from cyst.api.logic.access import AccessLevel
from cyst.api.utils.tag import Tag


class Service(ABC):
    """
    Service represents an abstraction of any notable process that is running on a node. Services exist in two flavors:
    passive and active. Active services can initiate request-response exchange and manage their own response processing.
    Passive services exist only as a descriptions of their properties and their responses are evaluated by the
    environment.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        The name of the service. This name is unique within a node and serves as a destination name for messages.

        :rtype: str
        """

    @property
    @abstractmethod
    def owner(self) -> str:
        """
        The identity of the owner of the service.

        :rtype: str
        """

    @property
    @abstractmethod
    def service_access_level(self) -> AccessLevel:
        """
        The access level with which the service can access resources of the node.

        :rtype: AccessLevel
        """

    @property
    @versionchanged(version="0.6.0", reason="Changed to return Optional ")
    @abstractmethod
    def passive_service(self) -> Optional['PassiveService']:
        """
        If the service is a passive one, this function gives access to its :class:`PassiveService` interface.

        :rtype: Optional[PassiveService]
        """

    @property
    @versionchanged(version="0.6.0", reason="Changed to return Optional ")
    @abstractmethod
    def active_service(self) -> Optional['ActiveService']:
        """
        If the service is an active one, this function gives access to its :class:`ActiveService` interface.

        :rtype: Optional[ActiveService]
        """


class ActiveService(ABC):
    """
    Active service is used to model any active actor within the simulation. Every agent that is being run in the
    simulation must implement this interface.
    """

    @abstractmethod
    async def run(self) -> None:
        """
        Starts the service. This function si automatically called by the environment at the initialization.
        """

    @abstractmethod
    async def process_message(self, message: Message) -> Tuple[bool, int]:
        """
        This function is called by the environment whenever a message arrives at the service.

        :param message: The message to process.
        :type message: Message

        :return: A tuple indicating, whether the processing was successful and eventually, how long the processing took.
            Depending on the role of the service, this return value results in different outcomes. For traffic,
            processors, i.e., services that process messages before they arrive into their final destination, returning
            (False, _) results in dropping of the message at that point. After returning (True, _) the messages is
            passed further. For destination services the resulting value currently does not play a role. Both types of
            services are expected to create and send an appropriate response. TODO: This should have the same result
            as the behavioral models and return a Response. That way, there can be a failsafe built-in.
        :rtype: Tuple[bool, int]
        """


@dataclass
class ActiveServiceDescription:
    """
    This is a description of an active service. It is used to register a new active service into the system.

    :param name: A name of the active service. It has to be unique among other registered services.
    :type name: str

    :param description: A short description of the service purpose and function.
    :type description: str

    :param creation_fn: A function that is able to create instances of the service. In addition to environment
        interfaces, the id and configuration dictionary is passed to service as well.
    :type creation_fn: Callable[[EnvironmentMessaging, EnvironmentResources, str, Optional[Dict[str, Any]]], ActiveService]
    """

    name: str
    description: str
    # TODO services are currently called with Dict[str, Any] for configuration. In the future, they should provide some
    #      information about their configuration
    creation_fn: Callable[['cyst.api.environment.messaging.EnvironmentMessaging',
                           'cyst.api.environment.resources.EnvironmentResources',
                           str, Optional[Dict[str, Any]]], ActiveService]


class ServiceState(IntEnum):
    """
    Service state represents the state of a service.

    Possible values:
        :INIT: A state that used when service is created and before environment was started.
        :RUNNING: A default state that is set when the environment has started.
        :PAUSED: A state in which service can't be accessed, but can be reactivated.
        :ZOMBIE: A state in which service can't be accessed and can't be reactivated.
    """
    INIT = 0
    RUNNING = 1
    PAUSED = 2
    ZOMBIE = 3


class PassiveService(Service, ABC):
    """
    Passive service is used to model an arbitrary service running on a node, which only passively reacts to the actions
    by other actors within a simulation.
    """

    @property
    @abstractmethod
    def version(self) -> VersionInfo:
        """
        The version of the service.

        :rtype: VersionInfo
        """

    @property
    @abstractmethod
    def state(self) -> ServiceState:
        """
        The state of the service.

        :rtype: ServiceState
        """

    @property
    @abstractmethod
    def tags(self) -> Set[Tag]:
        """
        Set of tags describing the service. It should provide high-level specification of service purpose. In the
        future, we expect to have a complete domain of services, from which the tags could be chosen. This will also
        be beneficial for automated generation of simulation scenarios.

        :rtype: Set[Tag]
        """

    @property
    @abstractmethod
    def enable_session(self) -> bool:
        """
        Returns, whether the service enables creation of sessions.

        :rtype: bool
        """

    @property
    @abstractmethod
    def session_access_level(self) -> AccessLevel:
        """
        Returns an access level of eventual sessions. This can be different from the access level of the service itself.
        For example ssh has elevated access level, but its sessions only have user level access.

        :rtype: AccessLevel
        """

    @property
    @abstractmethod
    def local(self) -> bool:
        """
        Returns whether the service is accessible from the network, or if the only way to address it is from the same
        node.

        :rtype: bool
        """



from typing import Dict

from cyst.api.utils.singleton import Singleton


class Counter(metaclass=Singleton):
    def __init__(self) -> None:
        self._counters: Dict[str, int] = {}

    def get(self, key: str) -> int:
        result = self._counters.get(key, None)
        if not result:
            result = 0

        self._counters[key] = result + 1
        return result



from copy import deepcopy
from typing import Any, List, Optional, Union, Type, Dict, Callable, Tuple
from netaddr import *

from cyst.api.environment.message import MessageType, Message, Request, Response, Status, Timeout, T, Resource, Signal, ComponentState
from cyst.api.host.service import ActiveService
from cyst.api.logic.access import Authorization, AuthenticationToken, AuthenticationTarget
from cyst.api.logic.action import Action
from cyst.api.logic.metadata import Metadata
from cyst.api.network.session import Session
from cyst.api.utils.counter import Counter

from cyst.platform.network.elements import Endpoint, Hop
from cyst.platform.network.session import SessionImpl


# TODO No repeated encapsulation of content yet
class Content:
    def __init__(self, encrypted_for=None, tokens=None, data=None):
        self._encrypted_for = encrypted_for
        self._tokens = tokens
        self._data = data


class MessageImpl(Message):
    def __init__(self, type: MessageType, origin: Endpoint = None, src_ip: IPAddress = None, dst_ip: IPAddress = None,
                 dst_service: str = "", session: Session = None,
                 auth: Optional[Union[Authorization, AuthenticationToken, AuthenticationTarget]] = None, force_id: int = -1, ttl: int = 64) -> None:

        super(MessageImpl, self).__init__()

        # Messages are globally indexed so that they can be ordered and are unique
        if force_id == -1:
            # Timeout is retaining its own counter, because it messes up the IDs if it is used extensively.
            self._id = Counter().get("timeout" if type == MessageType.TIMEOUT else "message")
        else:
            self._id = force_id

        self._type = type
        self._origin = origin
        self._src_ip = src_ip
        self._dst_ip = dst_ip
        self._src_service = ""
        self._dst_service = dst_service
        self._current = origin
        self._session = session
        self._auth = auth

        self._path: List[Hop] = []
        self._non_session_path: List[Hop] = []

        self._sent = False
        self._next_hop = None

        self._session_iterator = None
        self._in_session = False

        self._ttl = ttl

        self._metadata = None
        self._platform_specific: Dict[str, Any] = {}

    @property
    def id(self) -> int:
        return self._id

    @property
    def type(self) -> MessageType:
        return self._type

    @property
    def origin(self) -> Optional[Endpoint]:
        return self._origin

    def set_origin(self, value: Endpoint) -> None:
        self._origin = value
        # Setting origin automatically resets the value of current endpoint
        self._current = value

    @property
    def src_ip(self) -> Optional[IPAddress]:
        return self._src_ip

    def set_src_ip(self, value: IPAddress) -> None:
        self._src_ip = value

    @property
    def dst_ip(self) -> Optional[IPAddress]:
        return self._dst_ip

    def set_dst_ip(self, value: IPAddress) -> None:
        self._dst_ip = value

    @property
    def sent(self) -> bool:
        return self._sent

    @sent.setter
    def sent(self, value) -> None:
        self._sent = True

    @property
    def current(self) -> Optional[Endpoint]:
        return self._current


    @property
    def in_session(self) -> bool:
        return self._in_session

    def hop(self) -> None:
        self._current = self._next_hop

    @property
    def next_hop(self) -> Optional[Endpoint]:
        return self._next_hop

    # Next hop can be explicitly set by a switch, or can be taken from an active session
    def set_next_hop(self, origin_endpoint: Endpoint = None, destination_endpoint: Endpoint = None) -> None:
        if origin_endpoint and destination_endpoint:
            self._non_session_path.append(Hop(origin_endpoint, destination_endpoint))
            self._path.append(Hop(origin_endpoint, destination_endpoint))
            self._next_hop = destination_endpoint #type:ignore #MYPY: Should be fine
        else:
            # If it does not have a session then something is very wrong
            if not self._session:
                raise Exception("Message does not have a session to get next hop from")

            # Get a session iterator if the message did not enter session yet
            if not self._session_iterator:
                if self.type == MessageType.REQUEST:
                    self._session_iterator = SessionImpl.cast_from(self._session).forward_iterator #type:ignore #MYPY: Should be fine
                elif self.type == MessageType.RESPONSE:
                    self._session_iterator = SessionImpl.cast_from(self._session).reverse_iterator #type:ignore #MYPY: Should be fine
                else:
                    raise Exception("Attempting to send message other than request/response through session")

                self._in_session = True

            hop = next(self._session_iterator) #type: ignore #MYPY: probably fine
            src: Endpoint = hop.src
            self._next_hop = hop.dst

            """MYPY: something like this?             if self._current is not None and self._current.port == -1:
                self.set_origin(src)
"""

            if self._current.port == -1:
                self.set_origin(src)

            self._path.append(Hop(src, self._next_hop)) #MYPY: issue with hop, might currently return None, but should not based on annotation

            # If the next hop is one of the session's end, turn off session flag
            if (self.type == MessageType.REQUEST and self._next_hop.id == SessionImpl.cast_from(self._session).endpoint.id) or \
                (self.type == MessageType.RESPONSE and self._next_hop.id == SessionImpl.cast_from(self._session).startpoint.id):
                self._in_session = False #MYPY: if next hop really is None, then the call to id will fail

    # Can't really type it other then using string literal, because of dependency issues
    @property
    def session(self) -> Session:
        return self._session

    @session.setter
    def session(self, value: Session) -> None:
        self._session = value

    @property
    def auth(self) -> Optional[Union[Authorization, AuthenticationToken, AuthenticationTarget]]:
        return self._auth

    @auth.setter
    def auth(self, value: Union[Authorization, AuthenticationToken, AuthenticationTarget]) -> None:
        self._auth = value

    def __str__(self) -> str:
        result = "Message: [ID: {}, Type: {}, Origin: {}, Source: {}, Target: {}, Session: {}, Authorization: {}]"\
                 .format(self.id, self.type.name, self._origin, self.src_ip, self.dst_ip, self.session, self.auth)
        return result

    def __lt__(self, other) -> bool:
        if self.type.value != other.type.value:
            return self.type.value < other.type.value
        else:
            return self.id < other.id

    @property
    def path(self) -> List[Hop]:
        return self._path

    @property
    def non_session_path(self) -> List[Hop]:
        return self._non_session_path

    @property
    def src_service(self) -> str:
        return self._src_service

    @src_service.setter
    def src_service(self, value: str) -> None:
        self._src_service = value

    @property
    def dst_service(self):
        return self._dst_service

    @dst_service.setter
    def dst_service(self, value: str) -> None:
        self._dst_service = value

    @property
    def ttl(self):
        return self._ttl

    def decrease_ttl(self):
        self._ttl -= 1
        return self._ttl

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    def set_metadata(self, metadata: Metadata) -> None:
        self._metadata = metadata

    @property
    def platform_specific(self) -> Dict[str, Any]:
        return self._platform_specific

    @staticmethod
    def cast_from(o: Message) -> 'MessageImpl':
        if isinstance(o, MessageImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the Message interface")

    def cast_to(self, type: Type[T]) -> T:
        if isinstance(self, type):
            return self #MYPY: if this works, then probably ignore
        else:
            raise ValueError("Casting to a wrong derived type")


class RequestImpl(MessageImpl, Request):
    def __init__(self, dst_ip: Union[str, IPAddress], dst_service: str = "", action: Action = None,
                 session: Session = None, auth: Optional[Union[Authorization, AuthenticationToken, AuthenticationTarget]] = None,
                 original_request: Optional[Request] = None):

        if type(dst_ip) is str:
            dst_ip = IPAddress(dst_ip)

        _session = session
        if not _session and original_request:
            _session = original_request.session

        _auth = auth
        if not _auth and original_request:
            _auth = original_request.auth

        _origin = None
        if original_request:
            _origin = RequestImpl.cast_from(original_request).origin

        _src_ip = None
        if original_request:
            _src_ip = original_request.src_ip

        super(RequestImpl, self).__init__(MessageType.REQUEST, _origin, _src_ip, dst_ip, dst_service,
                                          session=_session, auth=_auth)

        if original_request:
            self.src_service = original_request.src_service
            self.platform_specific["caller_id"] = original_request.platform_specific["caller_id"]

        self._action = action

    @property
    def action(self) -> Action:
        return self._action

    @action.setter
    def action(self, value):
        self._action = value

    def __str__(self) -> str:
        result = "Request: [ID: {}, Type: {}, Origin: {}, Source: {}, Target: {}, Destination service: {}, Source service: {}, Action: {}, Session: {}, Authorization: {}]"\
                   .format(self.id, self.type.name, self._origin.ip if self.origin else None, self.src_ip, self.dst_ip, self.dst_service, self.src_service, self.action.id,
                           self.session, self.auth)
        return result


        """#MYPY: Maybe something liket this:
            def __str__(self) -> str:
        result = "Request: [ID: {}, Type: {}, Origin: {}, Source: {}, Target: {}, Destination service: {}, Source service: {}, Action: {}, Session: {}, Authorization: {}]"\
                   .format(self.id, self.type.name, self._origin.ip if self._origin is not None else None, self.src_ip, self.dst_ip, self.dst_service, self.src_service, self.action.id if self.action is not None else None,
                           self.session, self.auth)
        return result
        """



    @staticmethod
    def cast_from(o: Request) -> 'RequestImpl':
        if isinstance(o, RequestImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the Request interface")


class ResponseImpl(MessageImpl, Response):
    def __init__(self, request: MessageImpl, status: Status = None,
                 content: Any = None, session: Session = None,
                 auth: Optional[Union[Authorization, AuthenticationToken, AuthenticationTarget]] = None,
                 original_response: Optional[Response] = None) -> None:

        super(ResponseImpl, self).__init__(MessageType.RESPONSE, request.current, request.dst_ip, request.src_ip,
                                           session=session, auth=auth, force_id=request.id)

        self._status = status
        self._content = content
        if isinstance(request, Request):
            self._action = request.action
        else:
            raise RuntimeError("Attempting to create a response from non-request Message")
        # Response switches the source and destination services
        self._src_service = request.dst_service
        self._dst_service = request.src_service
        self._non_session_path = request._non_session_path
        self._path_index = len(self._non_session_path)
        self.set_origin(request.current)

        # Copy platform-specific information
        self._platform_specific = request.platform_specific

    def set_next_hop(self, origin_endpoint: Endpoint = None, destination_endpoint: Endpoint = None) -> None:
        # Traversing the pre-session connections is done by traversing back the non-session path
        # The rest is up to session management of message
        if self._path_index == 0 or self._in_session:
            return super(ResponseImpl, self).set_next_hop()

        self._path_index -= 1
        self._next_hop = self._non_session_path[self._path_index].src

    def __str__(self) -> str:
        result = "Response: [ID: {}, Type: {}, Origin: {}, Source: {}, Target: {}, Status: {}, Content: {}, Session: {}, Authorization: {}]"\
                   .format(self.id, self.type.name, self._origin.ip if self._origin else None, self.src_ip, self.dst_ip, self._status, self._content, self.session, self.auth)
        return result

    @property
    def action(self) -> Action:
        return self._action

    @property
    def status(self):
        return self._status

    @property
    def content(self):
        return self._content

    def set_in_session(self, value: bool = False) -> None:
        self._in_session = value

    @staticmethod
    def cast_from(o: Response) -> 'ResponseImpl':
        if isinstance(o, ResponseImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the Response interface")

    def __deepcopy__(self, memodict):
        cls = self.__class__
        result = cls.__new__(cls)
        memodict[id(self)] = result
        for k, v in self.__dict__.items():
            if k == "_content":
                setattr(result, "_content", str(v))
            else:
                setattr(result, k, deepcopy(v, memodict))
        return result

class TimeoutImpl(MessageImpl, Timeout):

    def __init__(self, callback: Union[ActiveService, Callable[[Message], Tuple[bool, int]]], start_time: float, duration: float, parameter: Optional[Any]):
        super(TimeoutImpl, self).__init__(MessageType.TIMEOUT)

        self._start_time = start_time
        self._duration = duration
        self._parameter = parameter
        self._callback = callback

    @property
    def start_time(self) -> float:
        return self._start_time

    @property
    def duration(self) -> float:
        return self._duration

    @property
    def parameter(self) -> Any:
        return self._parameter

    @property
    def callback(self) -> Callable[[Message], Tuple[bool, int]]:
        if isinstance(self._callback, ActiveService):
            return self._callback.process_message
        else:
            return self._callback


    def __str__(self) -> str:
        return "Timeout: [Start: {}, Duration: {}, Parameter: {}]".format(self._start_time, self._duration, self._parameter)

    @staticmethod
    def cast_from(o: Timeout) -> 'TimeoutImpl':
        if isinstance(o, TimeoutImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the Request interface")


class ResourceMessageImpl(MessageImpl, Resource):

    def __init__(self, path: str, status: Status, service: str, data: Optional[str]):
        super(ResourceMessageImpl, self).__init__(type=MessageType.RESOURCE, dst_service=service)

        self._path = path
        self._status = status
        self._data = data

    @property
    def path(self) -> str:
        return self._path

    @property
    def status(self) -> Status:
        return self._status

    @property
    def data(self) -> Optional[str]:
        return self._data


class SignalImpl(Signal):
    def __init__(self, signal_origin: str, state: ComponentState, effect_origin: str, effect_message: Optional[int],
                 effect_description: str, effect_parameters: Optional[Dict[str, Any]]):
        self._signal_origin = signal_origin
        self._state = state
        self._effect_origin = effect_origin
        self._effect_message = effect_message if effect_message else -1
        self._effect_description = effect_description
        self._effect_parameters = effect_parameters if effect_parameters else {}
        self._id = Counter().get("message")
        self._local_ip = IPAddress("127.0.0.1")
        self._metadata = Metadata()

    @property
    def signal_origin(self) -> str:
        return self._signal_origin

    @property
    def state(self) -> ComponentState:
        return self._state

    @property
    def effect_origin(self) -> str:
        return self._effect_origin

    @property
    def effect_message(self) -> Optional[int]:
        return self._effect_message

    @property
    def effect_description(self) -> str:
        return self._effect_description

    @property
    def effect_parameters(self) -> Dict[str, Any]:
        return self._effect_parameters

    @property
    def id(self) -> int:
        return self._id

    @property
    def type(self) -> MessageType:
        return MessageType.SIGNAL

    @property
    def src_ip(self) -> Optional[IPAddress]:
        return self._local_ip

    @property
    def dst_ip(self) -> Optional[IPAddress]:
        return self._local_ip

    @property
    def src_service(self) -> Optional[str]:
        return ""

    @property
    def dst_service(self) -> str:
        return ""

    @property
    def session(self) -> Optional[Session]:
        return None

    @property
    def auth(self) -> Optional[Union[Authorization, AuthenticationToken, AuthenticationTarget]]:
        return None

    @property
    def ttl(self) -> int:
        return 0

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    def set_metadata(self, metadata: Metadata) -> None:
        return

    @property
    def platform_specific(self) -> Dict[str, Any]:
        return {}

    def cast_to(self, type: Type[T]) -> T:
        raise ValueError("Cannot cast a signal to a message of a different type")



from cyst.api.environment.configuration import (
    EnvironmentConfiguration,
    GeneralConfiguration,
    NodeConfiguration,
    ServiceConfiguration,
    NetworkConfiguration,
    ExploitConfiguration,
    ActionConfiguration,
    AccessConfiguration,
    PhysicalConfiguration
)

class EnvironmentConfigurationImpl(EnvironmentConfiguration):
    def __init__(self, general: GeneralConfiguration, platform: EnvironmentConfiguration, action: ActionConfiguration,
                 exploit: ExploitConfiguration, physical: PhysicalConfiguration):
        self._general = general
        self._node = platform.node
        self._service = platform.service
        self._network = platform.network
        self._exploit = exploit
        self._action = action
        self._access = platform.access
        self._physical = physical

    @property
    def general(self) -> GeneralConfiguration:
        return self._general

    @property
    def node(self) -> NodeConfiguration:
        return self._node

    @property
    def service(self) -> ServiceConfiguration:
        return self._service

    @property
    def network(self) -> NetworkConfiguration:
        return self._network

    @property
    def exploit(self) -> ExploitConfiguration:
        return self._exploit

    @property
    def action(self) -> ActionConfiguration:
        return self._action

    @property
    def access(self) -> AccessConfiguration:
        return self._access

    @property
    def physical(self) -> PhysicalConfiguration:
        return self._physical



from datetime import datetime
from typing import List

from cyst.api.environment.physical import PhysicalAccess, PhysicalLocation, PhysicalConnection
from cyst.api.utils.duration import Duration


class PhysicalAccessImpl(PhysicalAccess):
    def __init__(self, identity: str, time_from: datetime | None = None, time_to: datetime | None = None):
        self._identity = identity
        self._time_from = time_from
        self._time_to = time_to

    @property
    def identity(self) -> str:
        return self._identity

    @property
    def time_from(self) -> datetime | None:
        return self._time_from

    @property
    def time_to(self) -> datetime | None:
        return self._time_to


class PhysicalLocationImpl(PhysicalLocation):
    def __init__(self):
        self._assets = []
        self._access = []

    @property
    def assets(self) -> List[str]:
        return self._assets

    @property
    def access(self) -> List[PhysicalAccess]:
        return self._access

    def add_asset(self, asset_id: str) -> None:
        if asset_id not in self._assets:
            self._assets.append(asset_id)

    def remove_asset(self, asset_id: str) -> None:
        if asset_id in self._assets:
            self._assets.remove(asset_id)

    def add_access(self, access: PhysicalAccess) -> None:
        self._access.append(access)

    def remove_access(self, access: PhysicalAccess) -> None:
        if access in self._access:
            self._access.remove(access)

class PhysicalConnectionImpl(PhysicalConnection):
    def __init__(self, origin: str, destination: str, travel_time: Duration):
        self._origin = origin
        self._destination = destination
        self._travel_time = travel_time

    @property
    def origin(self) -> str:
        return self._origin

    @property
    def destination(self) -> str:
        return self._destination

    @property
    def travel_time(self) -> Duration:
        return self._travel_time


from abc import ABC, abstractmethod


class Statistics(ABC):
    """
    Statistics class tracks various statistical information about simulation runs.
    """

    @property
    @abstractmethod
    def run_id(self) -> str:
        """
        Get the ID of the current run

        :rtype: str
        """

    @property
    @abstractmethod
    def configuration_id(self) -> str:
        """
        Get the ID of the configuration that is stored in the data store and that was used for the current run.

        :rtype: str
        """

    @property
    @abstractmethod
    def start_time_real(self) -> float:
        """
        Get the wall clock time when the current run started. The time is in the python time() format.

        :rtype: float
        """

    @property
    @abstractmethod
    def end_time_real(self) -> float:
        """
        Get the wall clock time when the current run was commited. The time is in the python time() format.

        :rtype: float
        """

    @property
    @abstractmethod
    def end_time_virtual(self) -> int:
        """
        Get the virtual time of the current run, i.e., the number of ticks. As the run always starts at 0, this function
        represents also the run duration.

        :rtype: int
        """



from __future__ import annotations

from heapq import heappush
from netaddr import IPAddress
from typing import TYPE_CHECKING, Optional, Any, Union, Dict, List

from cyst.api.environment.message import Request, Response, Status, Message, StatusValue, StatusOrigin, ComponentState, \
    Signal
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.host.service import ServiceState
from cyst.api.logic.access import Authorization, AuthenticationTarget, AuthenticationToken
from cyst.api.logic.action import Action, ActionType
from cyst.api.logic.metadata import Metadata
from cyst.api.network.firewall import Firewall
from cyst.api.network.session import Session
from cyst.api.utils.counter import Counter
from cyst.api.utils.duration import msecs

from cyst.platform.environment.message import RequestImpl, ResponseImpl, MessageImpl, MessageType, SignalImpl
from cyst.platform.host.service import ServiceImpl
# from cyst.platform.logic.action import ActionImpl
from cyst.platform.network.elements import Endpoint, InterfaceImpl
from cyst.platform.network.node import NodeImpl
from cyst.platform.network.session import SessionImpl

if TYPE_CHECKING:
    from cyst.platform.main import CYSTPlatform

class EnvironmentMessagingImpl(EnvironmentMessaging):
    def __init__(self, platform: CYSTPlatform):
        self._platform = platform

    def send_message(self, message: Message, delay: float = 0.0) -> None:
        # Messages with composite actions need to be processed via ActionManager
        # Logic:
        # if message.action.is_composite_action:
        #    self._env.composite_action_manager.process_composite_action(message)
        # else:
        #    # the rest
        m = MessageImpl.cast_from(message)
        _send_message(self._platform, m, delay)

    def create_request(self, dst_ip: Union[str, IPAddress], dst_service: str = "", action: Optional[Action] = None,
                       session: Optional[Session] = None,
                       auth: Optional[Union[Authorization, AuthenticationToken]] = None,
                       original_request: Optional[Request] = None) -> Request:
        return RequestImpl(dst_ip, dst_service, action, session, auth, original_request)

    def create_response(self, request: Request, status: Status, content: Optional[Any] = None,
                        session: Optional[Session] = None,
                        auth: Optional[Union[Authorization, AuthenticationTarget]] = None,
                        original_response: Optional[Response] = None):
        # Let's abuse the duck typing and "cast" Request to RequestImpl
        if isinstance(request, RequestImpl):
            response = ResponseImpl(request, status, content, session, auth, original_response)
            return response
        else:
            raise ValueError("Malformed request passed to create a response from")

    def create_signal(self, signal_origin: str, state: ComponentState, effect_origin: str,
                      effect_message: Optional[int] = None, effect_description: str = "",
                      effect_parameters: Optional[Dict[str, Any]] = None) -> Signal:
        return SignalImpl(signal_origin, state, effect_origin, effect_message, effect_description, effect_parameters)

    def open_session(self, request: Request, reverse_direction: bool = False) -> Session:
        return _open_session(self._platform, request, reverse_direction)

    async def message_hop(self, message: Message) -> None:
        await _message_hop(self._platform, message)

    async def message_process(self, message: Message) -> None:
        await _message_process(self._platform, message)

# ----------------------------------------------------------------------------------------------------------------------
# Free function implementations of the above class. It is being done this way to shut up the type checking and to
# overcome python's limitation on having a class implemented in multiple files.
def _open_session(self: CYSTPlatform, message: Message, reverse_direction: bool) -> Session:
    return self._network_configuration.create_session_from_message(message, reverse_direction)


def extract_metadata_action(action: Action, action_list: List[Action]):
    if not action.components:
        action_list.append(action)
    else:
        for c in action.components:
            extract_metadata_action(c, action_list)


def _send_message(self: CYSTPlatform, message: MessageImpl, delay: float = 0.0) -> None:
    # Shortcut for timeout
    if message.type == MessageType.TIMEOUT:
        heappush(self._message_queue, (self._time + delay, Counter().get("msg"), message))
        return

    if message.type == MessageType.REQUEST:
        # Set origin ID from platform-specific info
        caller_id: str = message.platform_specific["caller_id"]
        node_id, service_id = caller_id.split(".")

        message.set_origin(Endpoint(node_id, -1))
        message.src_service = service_id

    # set a first hop for a message
    source = self._network.get_node_by_id(message.origin.id)
    # Find a next hop for messages without one
    if source and not message.next_hop:
        # New request with session should follow the session first
        # Response should either follow newly established session, or route to session endpoint
        # TODO rearrange it to reflect changes in response set_next_hop handling
        if message.type == MessageType.REQUEST and message.session:
            message.set_next_hop()
            # Not a pretty thing, but I am not sure how to make it better
            # it = SessionImpl.cast_from(message.session).forward_iterator
            # hop = next(it)
            # port = hop.src.port
            # iface = source.interfaces[port]

            # If this works it is a proof that the entire routing must be reviewed
            message.set_src_ip(message.path[0].src.ip)
        elif message.type == MessageType.RESPONSE:
            if message.session and message.current == SessionImpl.cast_from(message.session).endpoint:
                # This is stupid, but it complains...
                if isinstance(message, ResponseImpl):
                    message.set_in_session(True)
            message.set_next_hop()
        # Others go to a gateway
        else:
            target = message.dst_ip
            localhost = IPAddress("127.0.0.1")

            # Shortcut for localhost request
            if target == localhost:
                message.set_src_ip(localhost)
                message.set_next_hop(Endpoint(message.origin.id, 0, localhost), Endpoint(message.origin.id, 0, localhost))

            else:
                gateway, port = source.gateway(target)
                if not gateway:
                    raise Exception("Could not send a message, no gateway to route it through.")

                iface = InterfaceImpl.cast_from(source.interfaces[port])
                message.set_src_ip(iface.ip)

                message.set_origin(Endpoint(source.id, port, iface.ip))

                # First sending is specific, because the current value is set to origin
                message.set_next_hop(message.origin, iface.endpoint)

    try:
        heappush(self._message_queue, (self._time + delay, Counter().get("msg"), message))
    except Exception as e:
        self._message_log.error(f"Error sending a message, reason: {e}")

# TODO: Time handling is still completely bonkers. If I send a message at time X and it should arrive at router at X+1
#       then any router responses are sent as from time X (log-wise).
async def _message_hop(self: CYSTPlatform, message: Message) -> None:
    message = MessageImpl.cast_from(message)

    # This is just for log messages
    message_type = str(message.type.name).lower()

    # Special processing on the first hop
    if not message.sent:
        message.sent = True
        self._message_log.debug(f"Sending a message: {str(message)}")

    # shortcut for wakeup messages
    # TODO: Commented out (origin may be problematic if not via send_message)
    # if message.type == MessageType.TIMEOUT:
    #     self._network.get_node_by_id(message.origin.id).process_message(message)  # MYPY: Node returned by get_node can be None
    #     return

    # Traffic processor are affecting request before they are even sent out (not on routers, as that would double
    # the processing)
    if message.type == MessageType.REQUEST:
        current_node: NodeImpl = self._network.get_node_by_id(message.current.id)
        if current_node.type != "Router":
            for processor in current_node.traffic_processors:
                result, delay = await processor.process_message(message)
                if not result:
                    return

    # Store it into the history
    # Move platform-specific-and-data-store-worthy information into the platform_specific attribute
    message.platform_specific["current_hop_ip"] = str(message.current.ip)
    message.platform_specific["current_hop_id"] = message.current.id
    message.platform_specific["next_hop_ip"] = str(message.next_hop.ip)
    message.platform_specific["next_hop_id"] = message.next_hop.id

    if self._message_storage:
        self._infrastructure.data_store.add_message(message)

    # Move message to a next hop
    message.hop()
    current_node: NodeImpl = self._network.get_node_by_id(message.current.id)  # MYPY: Get node can return None

    # If the message was not running around localhost, get the connection delay
    delay = 0
    if message.src_ip != message.dst_ip:
        connection = self.configuration.network.get_connections(current_node, message.current.port)[0]
        delay, result = connection.evaluate(message)
        if delay < 0:
            # TODO: message dropped, what to do? Maybe send early without processing
            pass

    message = MessageImpl.cast_from(message)
    # TODO: Parametrization of platform should be in one place and not be magic constants everywhere
    processing_time = max(msecs(20).to_float(), delay)

    # HACK: Because we want to enable actions to be able to target routers, we need to bypass the router processing
    #       if the message is at the end of its journey
    last_hop = message.dst_ip == message.current.ip  # MYPY: current can return None

    if not last_hop and current_node.type == "Router":
        result, delay = await current_node.process_message(message, processing_time)
        processing_time += delay
        if result:
            heappush(self._message_queue, (self._time + processing_time, Counter().get("msg"), message))

        return

    # Message has a session
    if message.session:
        local_processing = False
        # Message still in session, pass it along
        if message.in_session:
            message.set_next_hop()
            heappush(self._message_queue, (self._time + processing_time, Counter().get("msg"), message))
            return
        # The session ends in the current node
        elif message.session.endpoint.id == current_node.id or message.session.startpoint.id == current_node.id:  # MYPY: here on multiple line, session only has an end and start, not endpoint and startpoint
            # TODO bi-directional session complicate the situation soooo much
            end_port = None
            if message.session.endpoint.id == current_node.id:
                end_port = message.session.endpoint.port
            elif message.session.startpoint.id == current_node.id:
                end_port = message.session.startpoint.port

            # Check if the node is the final destination
            for iface in current_node.interfaces:
                if iface.index == end_port and iface.ip == message.dst_ip:  # MYPY: Interface does not have index
                    local_processing = True
                    break
            # It is not, this means the node was only a proxy to some other target
            if not local_processing:
                # Find a way to nearest switch
                gateway, port = current_node.gateway(message.dst_ip)  # MYPY: If this returns None, there is only one value and it will crash on unpacking it
                # ##################
                dest_node_endpoint = current_node.interfaces[port].endpoint  # MYPY: end vs endpoint
                dest_node = self._network.get_node_by_id(dest_node_endpoint.id)
                dest_node_ip = dest_node.interfaces[dest_node_endpoint.port].ip  # MYPY: dest_node can be None
                message.set_next_hop(Endpoint(current_node.id, port, current_node.interfaces[port].ip),
                                     Endpoint(dest_node_endpoint.id, dest_node_endpoint.port, dest_node_ip))
                # ##################
                self._message_log.debug(
                    f"Proxying a {message_type} to {message.dst_ip} via {message.next_hop.id} on a node {current_node.id}")
                heappush(self._message_queue, (self._time + processing_time, Counter().get("msg"), message))
                return

    # Message has to be processed locally
    heappush(self._execute_queue, (self._time + processing_time, Counter().get("msg"), message))


async def _message_process(self: CYSTPlatform, message: Message) -> None:
    message = MessageImpl.cast_from(message)
    processing_time = 0

    # This is just for log messages
    message_type = str(message.type.name).lower()
    current_node: NodeImpl = self._network.get_node_by_id(message.current.id)  # MYPY: Get node can return None

    self._message_log.debug(f"Processing a {message_type} on a node {current_node.id}. {message}")

    # Before a message reaches to services within, it is evaluated by all traffic processors
    # While they are returning true, everything is ok. Once they return false, the message processing stops
    # Traffic processors are free to send any reply as they see fit
    # TODO: Firewall does not return a response and currently we want it in some instances to return it and in
    #       some instances we don't. This is not a good situation.
    for processor in current_node.traffic_processors:
        # TODO Traffic processor responses are not correctly delayed
        result, delay = await processor.process_message(message)
        processing_time += delay
        if not result:
            return

    # Service is requested
    response = None
    if message.dst_service:
        # Check if the requested service exists on the current node
        if message.dst_service not in current_node.services:
            # There is a theoretical chance for response not finding dst service for responses, if e.g. attacker
            # shut down the service after firing request and before receiving the response. In such case the
            # error is silently dropped
            if message.type == MessageType.RESPONSE:
                return

            # TODO: Parametrization of platform
            processing_time += msecs(20).to_float()
            response = ResponseImpl(message, Status(StatusOrigin.NODE, StatusValue.ERROR),
                                    "Nonexistent service {} at node {}".format(message.dst_service, message.dst_ip),
                                    session=message.session, auth=message.auth)
            self._environment_messaging.send_message(response, processing_time)

        # Service exists and it is passive
        elif ServiceImpl.cast_from(current_node.services[message.dst_service]).passive:
            # Passive services just discard the responses and only process the requests
            if message.type == MessageType.RESPONSE:
                return

            if current_node.services[message.dst_service].passive_service.state != ServiceState.RUNNING:
                response = ResponseImpl(message, Status(StatusOrigin.NODE, StatusValue.ERROR),
                                        "Service {} at node {} is not running".format(message.dst_service, message.dst_ip),
                                        session=message.session, auth=message.auth)
                self._environment_messaging.send_message(response, processing_time)
            else:
                # Delay it by the time it took to process the last hop
                result, processing_time = self._platform_interface.execute_task(message, current_node.services[message.dst_service], current_node, processing_time)
                # delay, response = self._process_passive(message, current_node)
                # processing_time += delay
                # if response.status.origin == StatusOrigin.SYSTEM and response.status.value == StatusValue.ERROR:
                #    print("Could not process the request, unknown semantics.")
                # else:
                #    self._environment_messaging.send_message(response, processing_time)
        # Service exists and it is active
        else:
            # An active service does not necessarily produce Responses, so we should just move time
            # somehow and be done with it.
            # TODO How to move time? (I now know why, I need to investigate why)

            # The response is reaching its destination. If it is a part of a composite action, we defer the
            # processing to the composite action manager.
            # Logic:
            # - if message.is_composite_action_chain:
            #     # CompositeActionManager swallows messages belonging to the chain and only returns a processable
            #     # message after the final action is done or irrecoverable error is encountered
            #     message = CompositeActionManager.process(message)
            #
            #   if message:
            #        active_service.process_message()

            # TODO: CAM Ignored for now
            # if self._cam.is_composite(message.id):
            #     self._cam.incoming_message(message)
            # else:
            # result, delay = current_node.services[message.dst_service].active_service.process_message(message)
            result, delay = self._platform_interface.execute_task(message, current_node.services[message.dst_service], current_node, processing_time)

    # If no service is specified, it is a message to a node, but still, it is processed as a request for
    # passive service and processed with the interpreter
    # No service is specified
    else:
        # If there is response arriving without destination service, just drop it
        if message.type == MessageType.RESPONSE:
            return

        # If it is a request, then it is processed as a request for passive service and processed with the interpreter
        # Delay it by the time it took to process the last hop
        result, delay = self._platform_interface.execute_task(message, None, current_node, processing_time)
        # delay, response = self._process_passive(message, current_node) #MYPY: messageimpl vs request
        # processing_time += delay
        # if response.status.origin == StatusOrigin.SYSTEM and response.status.value == StatusValue.ERROR: #MYPY: same as above, response None?
        #    print("Could not process the request, unknown semantics.")
        # else:
        #    self._environment_messaging.send_message(response, processing_time)


from __future__ import annotations

from typing import TYPE_CHECKING, List, Union, Optional

from cyst.api.environment.configuration import NetworkConfiguration
from cyst.api.environment.message import Message, MessageType
from cyst.api.network.elements import Connection
from cyst.api.network.node import Node
from cyst.api.network.session import Session

from cyst.platform.environment.message import MessageImpl
from cyst.platform.host.service import ServiceImpl
from cyst.platform.network.elements import PortImpl
from cyst.platform.network.session import SessionImpl
from cyst.platform.network.node import NodeImpl

if TYPE_CHECKING:
    from cyst.platform.main import CYSTPlatform


class NetworkConfigurationImpl(NetworkConfiguration):
    def __init__(self, platform: CYSTPlatform):
        self._platform = platform

    def add_node(self, node: Node) -> None:
        return _add_node(self._platform, node)

    def add_connection(self, source: Node, target: Node, source_port_index: int = -1, target_port_index: int = -1,
                       net: str = "", connection: Optional[Connection] = None) -> Connection:
        return _add_connection(self._platform, source, target, source_port_index, target_port_index, net, connection)

    def get_connections(self, node: Node, port_index: Optional[int] = None) -> List[Connection]:
        return [ifc.connection for ifc in node.interfaces if ifc.connection and
                (not port_index or PortImpl.cast_from(ifc)._index == port_index)]

    def create_session(self, owner: str, waypoints: List[Union[str, Node]], src_service: Optional[str] = None,
                       dst_service: Optional[str] = None, parent: Optional[Session] = None,
                       defer: bool = False, reverse: bool = False, id: Optional[str] = None) -> Optional[Session]:
        return _create_session(self._platform, owner, waypoints, src_service, dst_service, parent, defer, reverse, id)

    def create_session_from_message(self, message: Message, reverse_direction: bool = False) -> Session:
        return _create_session_from_message(self._platform, message, reverse_direction)

    def append_session(self, original_session: Session, appended_session: Session) -> Session:
        return _append_session(self._platform, original_session, appended_session)


# ------------------------------------------------------------------------------------------------------------------
# NetworkConfiguration
def _add_node(self: CYSTPlatform, node: Node) -> None:
    self._network.add_node(NodeImpl.cast_from(node))


def _add_connection(self: CYSTPlatform, source: Node, target: Node, source_port_index: int = -1, target_port_index: int = -1,
                   net: str = "", connection: Connection = None) -> Connection:
    return self._network.add_connection(NodeImpl.cast_from(source), source_port_index, NodeImpl.cast_from(target),
                                        target_port_index, net, connection)


# TODO: Decide if we want to have service association a part of the session creation, or if we rather leave it
#       to service interface
def _create_session(self: CYSTPlatform, owner: str, waypoints: List[Union[str, Node]], src_service: Optional[str] = None,
                    dst_service: Optional[str] = None, parent: Optional[Session] = None, defer: bool = False,
                    reverse: bool = False, id: Optional[str] = None) -> Optional[Session]:

    if defer:
        self._sessions_to_add.append((owner, waypoints, src_service, dst_service, parent, reverse, id))
        return None
    else:
        session = self._network.create_session(owner, waypoints, src_service, dst_service, parent, reverse, id)
        if src_service or dst_service:
            if not src_service and dst_service:
                raise RuntimeError("Both or neither services must be specified during session creation.")

            src_node: Node
            if isinstance(waypoints[0], str):
                src_node = self._network.get_node_by_id(waypoints[0])
            else:
                src_node = waypoints[0]

            dst_node: Node
            if isinstance(waypoints[-1], str):
                dst_node = self._network.get_node_by_id(waypoints[-1])
            else:
                dst_node = waypoints[-1]

            ServiceImpl.cast_from(src_node.services[src_service]).add_session(session)
            ServiceImpl.cast_from(dst_node.services[dst_service]).add_session(session)
        return session


def _append_session(self: CYSTPlatform, original_session: Session, appended_session: Session) -> Session:
    original = SessionImpl.cast_from(original_session)
    appended = SessionImpl.cast_from(appended_session)

    src_service = original._src_service
    dst_service = appended._dst_service

    node = self._network.get_node_by_id(original.startpoint.id)
    session = SessionImpl(original.owner, original, appended.path_id, src_service, dst_service)

    ServiceImpl.cast_from(node.services[src_service]).add_session(session)

    return session


def _create_session_from_message(self: CYSTPlatform, message: Message, reverse_direction: bool = False) -> Session:
    message = MessageImpl.cast_from(message)

    if message.auth:
        owner = message.auth.identity
    else:
        owner = message.dst_service
    path = message.non_session_path
    parent = message.session

    # In case someone attempts to create another session in the endpoint of already provided session, just return
    # that session instead.
    # TODO: Document this behavior
    if not path:
        return parent

    # Source and destination services are taken from message and the session reference is inserted to both
    if message.type == MessageType.REQUEST:
        src_service = message.src_service
        dst_service = message.dst_service
    else:
        src_service = message.dst_service
        dst_service = message.src_service

    if reverse_direction:
        path = [hop.swap() for hop in reversed(path)]
        src_service, dst_service = dst_service, src_service

    session = SessionImpl(owner, parent, path, src_service, dst_service, self._network)

    if parent:
        p = SessionImpl.cast_from(parent)
        src_node = self._network.get_node_by_id(p.startpoint.id)
    else:
        src_node = self._network.get_node_by_id(path[0].src.id)
    dst_node = self._network.get_node_by_id(path[-1].dst.id)

    ServiceImpl.cast_from(src_node.services[src_service]).add_session(session)
    ServiceImpl.cast_from(dst_node.services[dst_service]).add_session(session)

    return session



from abc import abstractmethod, ABC
from datetime import datetime
from typing import List

from cyst.api.utils.duration import Duration


class PhysicalAccess(ABC):
    """
    Physical access specifies, which user can access given physical location in a given time. The access
    control works in a default deny mode, i.e., unless the user is cleared for access, the physical location is
    off-limits.
    """

    @property
    @abstractmethod
    def identity(self) -> str:
        """
        :return: A unique identifier of a user. An asterisk '*' denotes that this access applies to any user.
        """

    @property
    @abstractmethod
    def time_from(self) -> datetime | None:
        """
        :return: A time of a day (the date part is ignored) from which the access is granted.
        """

    @property
    @abstractmethod
    def time_to(self) -> datetime | None:
        """
        :return: A time of a day (the date part is ignored) to which the access is granted.
        """


class PhysicalLocation:
    """
    Physical location groups together physical assets, which will most often be Nodes and Routers. While Connections
    have their physical manifestations, they are not considered physical for this purpose.
    """

    @property
    @abstractmethod
    def assets(self) -> List[str]:
        """
        A list of IDs of items that are physically present at a given location.

        :return: A list of asset IDs in the configuration.
        """

    @property
    @abstractmethod
    def access(self) -> List[PhysicalAccess]:
        """
        :return: A list of physical access specifications.
        """


class PhysicalConnection(ABC):
    """
    Physical connection represents traversing between two locations. This can be anything from going through a door
    frame to boarding an intercontinental flight. Physical connections do not have access control, i.e., any user can
    attempt to cross the divide between two locations, but if they are denied access at the destination, it is assumed
    that the user will have to return back effectively wasting twice the time without accomplishing nothing.

    Physical connections are bidirectional, so there is no real difference between the origin and destination.

    N.B.: Situation when a user can have access to an infrastructure via wireless networks when travelling between
    locations is currently not considered.
    """

    @property
    @abstractmethod
    def origin(self) -> str:
        """
        :return: An id of the first location.
        """

    @property
    @abstractmethod
    def destination(self) -> str:
        """
        :return: An id of the second location.
        """

    @property
    @abstractmethod
    def travel_time(self) -> Duration:
        """
        :return: A time needed to traverse the physical connection.
        """



import inspect
import logging

from abc import ABC, abstractmethod
from typing import List, Tuple, Optional, Dict, Any, Union, Callable, Awaitable

from cyst.api.logic.action import Action
from cyst.api.logic.access import Authorization, AuthenticationToken
from cyst.api.environment.environment import EnvironmentMessaging
from cyst.api.environment.message import Request, Response, MessageType, Message
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.network.session import Session
from cyst.api.host.service import ActiveService, ActiveServiceDescription, Service


class ScriptedActorControl(ABC):
    @property
    @abstractmethod
    def sessions(self) -> Dict[str, Session]:
        pass

    @abstractmethod
    def execute_action(self, target: str, service: str, action: Action, session: Session | str | None = None,
                       auth: Optional[Union[Authorization, AuthenticationToken]] = None) -> None:
        pass

    @abstractmethod
    def get_last_message_type(self) -> Optional[MessageType]:
        pass

    @abstractmethod
    def get_last_request(self) -> Optional[Request]:
        pass

    @abstractmethod
    def get_last_response(self) -> Optional[Response]:
        pass

    @abstractmethod
    def set_run_callback(self, fn: Callable[[EnvironmentMessaging, EnvironmentResources], Awaitable[None]]):
        pass

    @abstractmethod
    def set_request_callback(self, fn: Callable[[EnvironmentMessaging, EnvironmentResources, Message], Tuple[bool, int]]):
        pass

    @abstractmethod
    def set_response_callback(self, fn: Callable[[EnvironmentMessaging, EnvironmentResources, Message], Tuple[bool, int]]):
        pass

    @abstractmethod
    def set_message_callback(self, message_type: MessageType, fn: Callable[[EnvironmentMessaging, EnvironmentResources, Message], Tuple[bool, int]]):
        pass


class ScriptedActor(ActiveService, ScriptedActorControl):
    def __init__(self, env: EnvironmentMessaging = None, res: EnvironmentResources = None, id: str = "", args: Optional[Dict[str, Any]] = None) -> None:
        self._id = id
        self._messaging = env
        self._resources = res
        self._responses = []
        self._requests = []
        self._run_callback = None
        self._response_callback = None
        self._request_callback = None
        self._last_message_type = None
        self._log = logging.getLogger("services.scripted_actor")
        self._callbacks = {}
        self._sessions = args.get("__sessions", {}) if args else {}

    # This Actor only runs given actions. No own initiative
    async def run(self):
        self._log.info("Launched a scripted Actor")
        if self._run_callback:
            await self._run_callback(self._messaging, self._resources)

    @property
    def sessions(self) -> Dict[str, Session]:
        return self._sessions

    def execute_action(self, target: str, service: str, action: Action, session: Session | str | None = None,
                       auth: Optional[Union[Authorization, AuthenticationToken]] = None) -> None:

        if isinstance(session, str):
            s = self._sessions.get(session, None)
            if not s:
                self._log.error(f"Attempted to use a session with id '{session}'. Service does not have it.")
                session = None
            else:
                session = s

        request = self._messaging.create_request(target, service, action, session=session, auth=auth)
        self._messaging.send_message(request)

    async def process_message(self, message: Message) -> Tuple[bool, int]:
        self._last_message_type = message.type

        type_map = {
            MessageType.REQUEST: "request",
            MessageType.RESPONSE: "response",
            MessageType.RESOURCE: "resource",
            MessageType.SIGNAL: "signal"
        }

        message_type = type_map[message.type]

        self._log.debug(f"Got a new {message_type} {message.id} : {str(message)}")

        if message.type == MessageType.REQUEST:
            self._requests.append(message)
        elif message.type == MessageType.RESPONSE:
            self._responses.append(message)

        if message.type in self._callbacks:
            return self._callbacks[message.type](self._messaging, self._resources, message)

        return True, 1

    def get_last_message_type(self) -> Optional[MessageType]:
        return self._last_message_type

    def get_last_request(self, count: int = 1) -> Optional[Union[Request, List[Request]]]:
        if not self._requests:
            return None
        if count == 1:
            return self._requests[-1]
        return self._requests[:-count-1:-1]

    def get_last_response(self, count: int = 1) -> Optional[Union[Response, List[Response]]]:
        if not self._responses:
            return None
        if count == 1:
            return self._responses[-1]
        return self._responses[:-count-1:-1]

    def set_request_callback(self, fn: Callable[[EnvironmentMessaging, EnvironmentResources, Message], Tuple[bool, int]]):
        self._callbacks[MessageType.REQUEST] = fn

    def set_response_callback(self, fn: Callable[[EnvironmentMessaging, EnvironmentResources, Message], Tuple[bool, int]]):
        self._callbacks[MessageType.RESPONSE] = fn

    def set_message_callback(self, message_type: MessageType, fn: Callable[[EnvironmentMessaging, EnvironmentResources, Message], Tuple[bool, int]]):
        self._callbacks[message_type] = fn

    def set_run_callback(self, fn: Callable[[EnvironmentMessaging, EnvironmentResources], Awaitable[Any]]):
        self._run_callback = fn

    @staticmethod
    def cast_from(o: Service) -> 'ScriptedActor':
        if o.active_service:
            # Had to do it step by step to shut up the validator
            service = o.active_service
            if isinstance(service, ScriptedActor):
                return service
            else:
                raise ValueError("Malformed underlying object passed with the Session interface")
        else:
            raise ValueError("Not an active service passed")


def create_actor(msg: EnvironmentMessaging, res: EnvironmentResources, id:str, args: Optional[Dict[str, Any]]) -> ActiveService:
    actor = ScriptedActor(msg, res, id, args)
    return actor


service_description = ActiveServiceDescription(
    "scripted_actor",
    "An actor that only performs given actions. No logic whatsoever.",
    create_actor
)
